# -*- coding: iso-8859-1 -*-
# transient.py
# Transient analysis
# Copyright 2006 Giuseppe Venturini

# This file is part of the ahkab simulator.
#
# Ahkab is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# Ahkab is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License v2
# along with ahkab.  If not, see <http://www.gnu.org/licenses/>.

""" This module provides the methods required to perform a transient analysis.

Our problem can be written as:

.. math::

    D \\cdot dx/dt + MNA \\cdot x + T_v(x) + T_t(t) + N = 0

We need:

    1. :math:`MNA`, the static Modified Nodal Analysis matrix,
    2. :math:`N`, constant DC term,
    3. :math:`T_v(x)`, the non-linear DC term
    4. :math:`T_t(t)`, the time variant term, time dependent-sources,
       to be evaluated at each time step,
    5. The dynamic :math:`D` matrix,
    6. a differentiation method to approximate :math:`dx/dt`.

"""

from __future__ import (unicode_literals, absolute_import,
                        division, print_function)

import sys
import imp

import numpy as np

from . import dc_analysis
from . import implicit_euler
from . import ticker
from . import options
from . import circuit
from . import printing
from . import utilities
from . import devices
from . import results

# differentiation methods, add them here
IMPLICIT_EULER = "IMPLICIT_EULER"
TRAP = "TRAP"
GEAR1 = "GEAR1"
GEAR2 = "GEAR2"
GEAR3 = "GEAR3"
GEAR4 = "GEAR4"
GEAR5 = "GEAR5"
GEAR6 = "GEAR6"

specs = {'tran':{'tokens':({
                          'label':'tstep',
                          'pos':0,
                          'type':float,
                          'needed':True,
                          'dest':'tstep',
                          'default':None
                         },
                         {
                          'label':'tstop',
                          'pos':1,
                          'type':float,
                          'needed':True,
                          'dest':'tstop',
                          'default':None
                         },
                         {
                          'label':'tstart',
                          'pos':None,
                          'type':float,
                          'needed':False,
                          'dest':'tstart',
                          'default':0
                         },
                         {
                          'label':'uic',
                          'pos':2,
                          'type':float,
                          'needed':False,
                          'dest':'uic',
                          'default':0
                         },
                         {
                          'label':'ic_label',
                          'pos':None,
                          'type':str,
                          'needed':False,
                          'dest':'x0',
                          'default':0
                         },
                         {
                          'label':'method',
                          'pos':None,
                          'type':str,
                          'needed':False,
                          'dest':'method',
                          'default':None
                         }
                        )
               }
           }


def transient_analysis(circ, tstart, tstep, tstop, method=options.default_tran_method, use_step_control=True, x0=None,
                       mna=None, N=None, D=None, outfile="stdout", return_req_dict=None, verbose=3):
    """Performs a transient analysis of the circuit described by circ.

    Parameters:
    circ: circuit instance to be simulated.
    tstart: start value. Better leave this to zero.
    tstep: the maximum step to be allowed during simulation or
    tstop: stop value for simulation
    method: differentiation method: 'TRAP' (default) or 'IMPLICIT_EULER' or 'GEARx' with x=1..6
    use_step_control: the LTE will be calculated and the step adjusted. default: True
    x0: the starting point, the solution at t=tstart (defaults to None, will be set to the OP)
    mna, N, D: MNA matrices, defaulting to None, for big circuits, reusing matrices saves time
    outfile: filename, the results will be written to this file. "stdout" means print out.
    return_req_dict:  to be documented
    verbose: verbosity level from 0 (silent) to 6 (very verbose).

    """
    if outfile == "stdout":
        verbose = 0
    _debug = False
    if options.transient_no_step_control:
        use_step_control = False
    if _debug:
        print_step_and_lte = True
    else:
        print_step_and_lte = False

    method = method.upper() if method is not None else options.default_tran_method
    HMAX = tstep

    #check parameters
    if tstart > tstop:
        printing.print_general_error("tstart > tstop")
        sys.exit(1)
    if tstep < 0:
        printing.print_general_error("tstep < 0")
        sys.exit(1)

    if verbose > 4:
        tmpstr = "Vea = %g Ver = %g Iea = %g Ier = %g max_time_iter = %g HMIN = %g" % \
        (options.vea, options.ver, options.iea, options.ier, options.transient_max_time_iter, options.hmin)
        printing.print_info_line((tmpstr, 5), verbose)

    locked_nodes = circ.get_locked_nodes()

    if print_step_and_lte:
        flte = open("step_and_lte.graph", "w")
        flte.write("#T\tStep\tLTE\n")

    printing.print_info_line(("Starting transient analysis: ", 3), verbose)
    printing.print_info_line(("Selected method: %s" % (method,), 3), verbose)
    #It's a good idea to call transient with prebuilt MNA and N matrix
    #the analysis will be slightly faster (long netlists).
    if mna is None or N is None:
        (mna, N) = dc_analysis.generate_mna_and_N(circ, verbose=verbose)
        mna = utilities.remove_row_and_col(mna)
        N = utilities.remove_row(N, rrow=0)
    elif not mna.shape[0] == N.shape[0]:
        printing.print_general_error("mna matrix and N vector have different number of columns.")
        sys.exit(0)
    if D is None:
        # if you do more than one tran analysis, output streams should be changed...
        # this needs to be fixed
        D = generate_D(circ, (mna.shape[0], mna.shape[0]))
        D = utilities.remove_row_and_col(D)

    # setup x0
    if x0 is None:
        printing.print_info_line(("Generating x(t=%g) = 0" % (tstart,), 5), verbose)
        x0 = np.zeros((mna.shape[0], 1))
        opsol =  results.op_solution(x=x0, error=x0, circ=circ, outfile=None)
    else:
        if isinstance(x0, results.op_solution):
            opsol = x0
            x0 = x0.asarray()
        else:
            opsol =  results.op_solution(x=x0, error=np.zeros((mna.shape[0], 1)), circ=circ, outfile=None)
        printing.print_info_line(("Using the supplied op as x(t=%g)." % (tstart,), 5), verbose)

    if verbose > 4:
        print("x0:")
        opsol.print_short()

    # setup the df method
    printing.print_info_line(("Selecting the appropriate DF ("+method+")... ", 5), verbose, print_nl=False)
    if method == IMPLICIT_EULER:
        from . import implicit_euler as df
    elif method == TRAP:
        from . import trap as df
    elif method == GEAR1:
        from . import gear as df
        df.order = 1
    elif method == GEAR2:
        from . import gear as df
        df.order = 2
    elif method == GEAR3:
        from . import gear as df
        df.order = 3
    elif method == GEAR4:
        from . import gear as df
        df.order = 4
    elif method == GEAR5:
        from . import gear as df
        df.order = 5
    elif method == GEAR6:
        from . import gear as df
        df.order = 6
    else:
        df = import_custom_df_module(method, print_out=(outfile != "stdout"))
        # df is none if module is not found

    if df is None:
        sys.exit(23)

    if not df.has_ff() and use_step_control:
        printing.print_warning("The chosen DF does not support step control. Turning off the feature.")
        use_step_control = False
        #use_aposteriori_step_control = False

    printing.print_info_line(("done.", 5), verbose)

    # setup the data buffer
    # if you use the step control, the buffer has to be one point longer.
    # That's because the excess point is used by a FF in the df module to predict the next value.
    printing.print_info_line(("Setting up the buffer... ", 5), verbose, print_nl=False)
    ((max_x, max_dx), (pmax_x, pmax_dx)) = df.get_required_values()
    if max_x is None and max_dx is None:
        printing.print_general_error("df doesn't need any value?")
        sys.exit(1)
    if use_step_control:
        buffer_len = 0
        for mx in (max_x, max_dx, pmax_x, pmax_dx):
            if mx is not None:
                buffer_len = max(buffer_len, mx)
        buffer_len += 1
        thebuffer = dfbuffer(length=buffer_len, width=3)
    else:
        thebuffer = dfbuffer(length=max(max_x, max_dx) + 1, width=3)
    thebuffer.add((tstart, x0, None)) #setup the first values
    printing.print_info_line(("done.", 5), verbose) #FIXME

    #setup the output buffer
    if return_req_dict:
        output_buffer = dfbuffer(length=return_req_dict["points"], width=1)
        output_buffer.add((x0,))
    else:
        output_buffer = None

    # import implicit_euler to be used in the first iterations
    # this is because we don't have any dx when we start, nor any past point value
    if (max_x is not None and max_x > 0) or max_dx is not None:
        from . import implicit_euler
        first_iterations_number = max_x if max_x is not None else 1
        first_iterations_number = max( first_iterations_number, max_dx+1) \
                                  if max_dx is not None else first_iterations_number
    else:
        first_iterations_number = 0

    printing.print_info_line(("MNA (reduced):", 5), verbose)
    printing.print_info_line((mna, 5), verbose)
    printing.print_info_line(("D (reduced):", 5), verbose)
    printing.print_info_line((D, 5), verbose)

    # setup the initial values to start the iteration:
    x = None
    time = tstart
    nv = circ.get_nodes_number()

    Gmin_matrix = dc_analysis.build_gmin_matrix(circ, options.gmin, mna.shape[0], verbose)

    # lo step viene generato automaticamente, ma non superare mai quello fornito.
    if use_step_control:
        #tstep = min((tstop-tstart)/9999.0, HMAX, 100.0 * options.hmin)
        tstep = min((tstop-tstart)/9999.0, HMAX)
    printing.print_info_line(("Initial step: %g"% (tstep,), 5), verbose)

    if max_dx is None:
        max_dx_plus_1 = None
    else:
        max_dx_plus_1 = max_dx +1
    if pmax_dx is None:
        pmax_dx_plus_1 = None
    else:
        pmax_dx_plus_1 = pmax_dx +1

    # setup error vectors
    aerror = np.zeros((x0.shape[0], 1))
    aerror[:nv-1, 0] = options.vea
    aerror[nv-1:, 0] = options.vea
    rerror = np.zeros((x0.shape[0], 1))
    rerror[:nv-1, 0] = options.ver
    rerror[nv-1:, 0] = options.ier

    iter_n = 0  # contatore d'iterazione
    # when to start predicting the next point
    start_pred_iter = max(*[i for i in (0, pmax_x, pmax_dx_plus_1) if i is not None])
    lte = None
    sol = results.tran_solution(circ, tstart, tstop, op=x0, method=method, outfile=outfile)
    printing.print_info_line(("Solving... ", 3), verbose, print_nl=False)
    tick = ticker.ticker(increments_for_step=1)
    tick.display(verbose > 1)
    while time < tstop:
        if iter_n < first_iterations_number:
            x_coeff, const, x_lte_coeff, prediction, pred_lte_coeff = \
            implicit_euler.get_df((thebuffer.get_df_vector()[0],), tstep, \
            predict=(use_step_control and iter_n >= start_pred_iter))
        else:
            x_coeff, const, x_lte_coeff, prediction, pred_lte_coeff = \
                df.get_df(thebuffer.get_df_vector(), tstep,
                          predict=(use_step_control and
                                   iter_n >= start_pred_iter)
                         )

        if options.transient_prediction_as_x0 and use_step_control and prediction is not None:
            x0 = prediction
        elif x is not None:
            x0 = x

        x1, error, solved, n_iter = dc_analysis.dc_solve(
                                                     mna=(mna + np.multiply(x_coeff, D)),
                                                     Ndc=N,  Ntran=np.dot(D, const), circ=circ,
                                                     Gmin=Gmin_matrix, x0=x0,
                                                     time=(time + tstep),
                                                     locked_nodes=locked_nodes,
                                                     MAXIT=options.transient_max_nr_iter,
                                                     verbose=0
                                                     )

        if solved:
            old_step = tstep #we will modify it, if we're using step control otherwise it's the same
            # step control (yeah)
            if use_step_control:
                if x_lte_coeff is not None and pred_lte_coeff is not None and prediction is not None:
                    # this is the Local Truncation Error :)
                    lte = abs((x_lte_coeff / (pred_lte_coeff - x_lte_coeff)) * (prediction - x1))
                    # it should NEVER happen that new_step > 2*tstep, for stability
                    new_step_coeff = 2
                    for index in range(x.shape[0]):
                        if lte[index, 0] != 0:
                            new_value = ((aerror[index, 0] + rerror[index, 0]*abs(x[index, 0])) / lte[index, 0]) \
                            ** (1.0 / (df.order+1))
                            if new_value < new_step_coeff:
                                new_step_coeff = new_value
                            #print new_value
                    new_step = tstep * new_step_coeff
                    if (options.transient_use_aposteriori_step_control and
                        new_step_coeff <
                        options.transient_aposteriori_step_threshold):
                        #don't recalculate a x for a small change
                        tstep = check_step(new_step, time, tstop, HMAX)
                        #print "Apost. (reducing) step = "+str(tstep)
                        continue
                    tstep = check_step(new_step, time, tstop, HMAX) # used in the next iteration
                    #print "Apriori tstep = "+str(tstep)
                else:
                    #print "LTE not calculated."
                    lte = None
            if print_step_and_lte and lte is not None:
                #if you wish to look at the step. We print just a lte
                flte.write(str(time)+"\t"+str(old_step)+"\t"+str(lte.max())+"\n")
            # if we get here, either aposteriori_step_control is
            # disabled, or it's enabled and the error is small
            # enough. Anyway, the result is GOOD, STORE IT.
            time = time + old_step
            x = x1
            iter_n = iter_n + 1
            sol.add_line(time, x)

            dxdt = np.multiply(x_coeff, x) + const
            thebuffer.add((time, x, dxdt))
            if output_buffer is not None:
                output_buffer.add((x, ))
            tick.step()
        else:
            # If we get here, Newton failed to converge. We need to reduce the step...
            if use_step_control:
                tstep = tstep/5.0
                tstep = check_step(tstep, time, tstop, HMAX)
                printing.print_info_line(("At %g s reducing step: %g s (convergence failed)" % (time, tstep), 5), verbose)
            else: #we can't reduce the step
                printing.print_general_error("Can't converge with step "+str(tstep)+".")
                printing.print_general_error("Try setting --t-max-nr to a higher value or set step to a lower one.")
                solved = False
                break
        if options.transient_max_time_iter and iter_n == options.transient_max_time_iter:
            printing.print_general_error("MAX_TIME_ITER exceeded ("+str(options.transient_max_time_iter)+"), iteration halted.")
            solved = False
            break

    if print_step_and_lte:
        flte.close()

    tick.hide(verbose > 1)

    if solved:
        printing.print_info_line(("done.", 3), verbose)
        printing.print_info_line(("Average time step: %g" % ((tstop - tstart)/iter_n,), 3), verbose)

        if output_buffer:
            ret_value = output_buffer.get_as_matrix()
        else:
            ret_value = sol
    else:
        print("failed.")
        ret_value =  None

    return ret_value

def check_step(tstep, time, tstop, HMAX):
    """Checks the step for several common issues and corrects them.

    The following problems are checked:

    - the step must be shorter than ``HMAX``. In the context of a transient
      analysis, that usually is the time step provided by the user,
    - the step must be equal or shorter than the simulation time left (ie stop
      time minus current time),
    - the step must be longer than ``options.hmin``, the minimum allowable time
      step. If the step goes below this value, convergence problems due to
      machine precision will occur. Typically when this happens, we halt the
      simulation.

    **Parameters:**

    tstep : float
        The time step, in second, that needs to be checked.
    time : float
        The current simulation time.
    tstop : float
        The time at which the simulation ends.
    HMAX : float
        The maximum allowable time step.

    **Returns:**

    tstep : float
        The step provided if it passes the tests, a *shortened* step otherwise.

    :raises ValueError: When the step is shorter than ``option.hmin``.

    """
    if tstep > HMAX:
        tstep = HMAX
    if tstop - time < tstep:
        tstep = tstop - time
    elif tstep < options.hmin:
        printing.print_general_error("Step size too small: "+str(tstep))
        raise ValueError("Step size too small")
    return tstep

def generate_D(circ, shape):
    """Generates the D matrix

    For every time t, the D matrix is used (elsewhere) to solve the following system:

    .. math::

        D dx/dt + MNA x + N + T(x) = 0

    It's easy to set up the KCL law for the voltage unknowns, capacitors
    introduce stamps just like resistors do in the MNA and we know that row 1
    refers to node 1, row 2 refers to node 2, and so on

    Inductors generate, together with voltage sources, ccvs, vcvs, a additional
    line in the MNA matrix, and hence in D too.

    The current flowing through the device gets added to the x vector.

    In the case of an inductors, we have:

    .. math::

        V(n_1) - V(n_2) - V_L = 0

    Where:

    .. math::

        V_L = L dI/dt

    That's 0 (zero) in DC analysis, but not in transient analysis, where it
    needs to be differentiated.

    To understand on which line does the inductor's L*dI/dt go, we use the order
    of the elements in `circuit`: first are all voltage lines, then the current
    ones in the same order of the elements that introduce
    them. Therefore, we need to access the circuit (`circ`).

    **Parameters:**

    circ : circuit instance
        The circuit instance for which the :math:`D` matrix is computed.

    shape : tuple of ints
        The shape of the *reduced* :math:`MNA` matrix, D will be of the same
        shape.

    **Returns:**

    D : ndarray
        The *unreduced* D matrix.
    """
    D = np.zeros((shape[0]+1, shape[1]+1))
    nv = circ.get_nodes_number()# - 1
    i_eq = 0 #each time we find a vsource or vcvs or ccvs, we'll add one to this.
    for elem in circ:
        if circuit.is_elem_voltage_defined(elem) and not isinstance(elem, devices.Inductor):
            i_eq = i_eq + 1
        elif isinstance(elem, devices.Capacitor):
            n1 = elem.n1
            n2 = elem.n2
            D[n1, n1] = D[n1, n1] + elem.value
            D[n1, n2] = D[n1, n2] - elem.value
            D[n2, n2] = D[n2, n2] + elem.value
            D[n2, n1] = D[n2, n1] - elem.value
        elif isinstance(elem, devices.Inductor):
            D[ nv + i_eq, nv + i_eq ] = -1 * elem.value
            # Mutual inductors (coupled inductors)
            # need to add a -M dI/dt where I is the current in the OTHER inductor.
            if len(elem.coupling_devices):
                for cd in elem.coupling_devices:
                    # get id+descr of the other inductor (eg. "L32")
                    other_id_wdescr = cd.get_other_inductor(elem.part_id)
                    # find its index to know which column corresponds to its current
                    other_index = circ.find_vde_index(other_id_wdescr, verbose=0)
                    # add the term.
                    D[ nv + i_eq, nv + other_index ] += -1 * cd.M
            # carry on as usual
            i_eq = i_eq + 1

    if options.cmin > 0:
        cmin_mat = np.eye(shape[0]+1-i_eq)
        cmin_mat[0, 1:] = 1
        cmin_mat[1:, 0] = 1
        cmin_mat[0, 0] = cmin_mat.shape[0]-1
        if i_eq:
            D[:-i_eq, :-i_eq] += options.cmin*cmin_mat
        else:
            D += options.cmin*cmin_mat
    return D

class dfbuffer:
    """This is a LIFO buffer with a method to read it all without deleting the elements.

    Newer entries are added on top of the buffer. It checks the size of the
    added elements, to be sure they are of the same size.

    **Parameters:**

    length : int
        The length of the buffer. Samples are added at index ``0``, shifting all
        the previous samples back to higher indices. Samples at an index equal
        to ``length`` (or higher) are discarded without notice.
    width : int
        The width of the buffer, every time :func:`add` is called, it must be to
        add a tuple of the same length as this parameter.
    """
    _the_real_buffer = None
    _length = 0
    _width  = 0

    def __init__(self, length, width):
        self._the_real_buffer = []
        self._length = length
        self._width = width

    def add(self, atuple):
        """Add a new data point to the buffer.

        **Parameters:**

        atuple : tuple of floats
            The data point to be added. Notice that the length of the tuple must
            agree with the width of the buffer.

        :raises ValueError: if the provided tuple and the buffer width do not
        match.

        """
        if not len(atuple) == self._width:
            raise ValueError("Attempted to add a element of wrong size to the" +
                             "LIFO buffer.")
        self._the_real_buffer.insert(0, atuple)
        if len(self._the_real_buffer) > self._length:
            self._the_real_buffer = self._the_real_buffer[:self._length]

    def get_df_vector(self):
        """Read out the contents of the buffer, without any modification

        This method, in the context of a transient analysis, returns a vector
        suitable for a differentiation formula.

        **Returns:**

        vec : list of tuples
            a list of tuples, each tuple being composed of ``width`` floats. In
            the context of a transient analysis, the list (or vector) conforms
            to the specification of the differentiation formulae.
            That is, the simulator stores in the buffer a list similar to::

                [[time(n), x(n), dx(n)], [time(n-1), x(n-1), dx(n-1)], ...]

        """
        return self._the_real_buffer

    def isready(self):
        """This shouldn't be used to determine if the buffer has enough points to
        use the df _if_ you use the step control.
        In that case, it holds even the points required for the FF.
        """
        if len(self._the_real_buffer) == self._length:
            return True
        else:
            return False

    def get_as_matrix(self):
        for vindex in range(self._width):
            for index in range(len(self._the_real_buffer)):
                if index == 0:
                    single_matrix = self._the_real_buffer[index][vindex]
                else:
                    single_matrix = np.concatenate((self._the_real_buffer[index][vindex], single_matrix), axis=0)
            if vindex == 0:
                complete_matrix = single_matrix
            else:
                complete_matrix = np.concatenate((complete_matrix, single_matrix), axis=1)
        return complete_matrix

def import_custom_df_module(method, print_out):
    """Imports a module that implements differentiation formula through imp.load_module
    Parameters:
    method: a string, the name of the df method module
    print_out: print to stdout some verbose messages

    Returns:
    The df module or None if the module is not found.
    """
    try:
        df = imp.load_module(imp.find_module(method.lower()))
        if print_out:
            print("Custom df module "+method.lower()+" loaded.")
    except:
        printing.print_general_error("Unrecognized method: "+method.lower()+".")
        df = None

    return df



# -*- coding: iso-8859-1 -*-
# netlist_parser.py
# Netlist parser module
# Copyright 2006 Giuseppe Venturini

# This file is part of the ahkab simulator.
#
# Ahkab is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# Ahkab is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License v2
# along with ahkab.  If not, see <http://www.gnu.org/licenses/>.

"""Parse spice-like netlist files and generate circuits instances.

The syntax is explained in :doc:`help/Netlist-Syntax` and it's based on [#1]_
whenever possible.

.. [#1] http://newton.ex.ac.uk/teaching/CDHW/Electronics2/userguide/

Introduction
------------

This module has one main circuit that is expected to be useful to the end user:
:func:`parse_circuit`, which encapsulates parsing a netlist file and returns the
circuit, the simulation objects and the post-processing directives (such as
plotting instructions).

Additionally, the module provides utility functions related to parsing, among
which the end user may be interested in the :func:`convert` function, which allows
converting from SPICE-like representations of floats, booleans and strings to
their Python representations.

The last type of functions in the module are utility functions to go through the
netlist files and remove comments.

Except for the aforementioned functions, the rest seem to be more suitable for
developers than end users.

Overview
--------

Function for parsing
====================

.. autosummary::
    parse_circuit
    main_netlist_parser
    parse_elem_resistor
    parse_elem_capacitor
    parse_elem_inductor
    parse_elem_inductor_coupling
    parse_elem_vsource
    parse_elem_isource
    parse_elem_diode
    parse_elem_mos
    parse_elem_vcvs
    parse_elem_vccs
    parse_elem_cccs
    parse_elem_ccvs
    parse_elem_switch
    parse_elem_user_defined
    parse_models
    parse_time_function
    parse_postproc
    parse_ics
    parse_analysis
    parse_single_analysis
    parse_temp_directive
    parse_param_value_from_string
    parse_ic_directive
    parse_sub_declaration
    parse_sub_instance
    parse_include_directive

Utility functions for conversions
=================================

.. autosummary::
    convert
    convert_units
    convert_boolean

Utility functions for file/txt handling
=======================================

.. autosummary::
    join_lines
    is_valid_value_param_string
    get_next_file_and_close_current


Module reference
----------------
"""

from __future__ import (unicode_literals, absolute_import,
                        division, print_function)

import sys
import imp
import math
import copy
import os

from . import circuit
from . import dc_analysis
from . import devices
from . import diode
from . import mosq
from . import ekv
from . import switch
from . import printing
from . import utilities
from . import plotting
from . import options

# analyses syntax
from .dc_analysis import specs as dc_spec
from .ac import specs as ac_spec
from .transient import specs as tran_spec
from .pss import specs as pss_spec
from .py3compat import StringIO
from .pz import specs as pz_specs
from .symbolic import specs as symbolic_spec
from .time_functions import time_fun_specs
from .time_functions import sin, pulse, exp, sffm, am

from .fourier import specs as fft_specs

specs = {}
for i in dc_spec, ac_spec, tran_spec, pss_spec, symbolic_spec, pz_specs:
    specs.update(i)

time_functions = {}
for i in sin, pulse, exp, sffm, am:
    time_functions.update({i.__name__:i})

def parse_circuit(filename, read_netlist_from_stdin=False):
    """Parse a SPICE-like netlist

    Directives are collected in lists and returned too, except for
    subcircuits, those are added to circuit.subckts_dict.

    **Returns:**

    (circuit_instance, analyses, plotting directives)
    """
    # Lots of differences with spice's syntax:
    # Support for alphanumeric node names, but the ref has to be 0. always
    # .end is not required, but if is used anything following it is ignored
    # many others, see doc.

    circ = circuit.Circuit(title="", filename=filename)

    if not read_netlist_from_stdin:
        ffile = open(filename, "r")
    else:
        buf = ""
        for aline in sys.stdin:
            buf += aline + "\n"
        ffile = StringIO(buf)

    file_list = [(ffile, "unknown", not read_netlist_from_stdin)]
    netlist_wd = os.path.split(filename)[0]
    file_index = 0
    directives = []
    model_directives = []
    postproc = []
    subckts_list_temp = []
    netlist_lines = []
    current_subckt_temp = []
    within_subckt = False
    line_n = 0

    try:
        while ffile is not None:
            while True:
                line = ffile.readline()
                if len(line) == 0:
                    break  # check for EOF
                line_n = line_n + 1
                line = line.strip().lower()
                if line_n == 1:
                    # the first line is always the title
                    circ.title = line
                    continue
                elif len(line) == 0:
                    continue  # empty line is really empty after strip()
                line = join_lines(ffile, line)
                if line[0] == "*":  # comments start with *
                    continue

                # directives are grouped together and evaluated after
                # we have the whole circuit.
                # subcircuits are grouped too, but processed first
                if line[0] == ".":
                    line_elements = line.split()
                    if line_elements[0] == '.subckt':
                        if within_subckt:
                            raise NetlistParseError("nested subcircuit declaration detected")
                        current_subckt_temp = current_subckt_temp + \
                            [(line, line_n)]
                        within_subckt = True
                    elif line_elements[0] == '.ends':
                        if not within_subckt:
                            raise NetlistParseError(".ENDS outside of .subckt")
                        within_subckt = False
                        subckts_list_temp.append(current_subckt_temp)
                        current_subckt_temp = []
                    elif line_elements[0] == '.include':
                        file_list.append(
                            parse_include_directive(line, netlist_wd))
                    elif line_elements[0] == ".end":
                        break
                    elif line_elements[0] == ".plot":
                        postproc.append((line, line_n))
                    elif line_elements[0] == '.four':
                        postproc.append((line, line_n))
                    elif line_elements[0] == '.fft':
                        postproc.append((line, line_n))
                    elif line_elements[0] == ".model":
                        model_directives.append((line, line_n))
                    else:
                        directives.append((line, line_n))
                    continue

                if within_subckt:
                    current_subckt_temp  = current_subckt_temp + \
                        [(line, line_n)]
                else:
                    netlist_lines = netlist_lines + [(line, line_n)]
            if within_subckt:
                raise NetlistParseError(".ends not found")

            file_index = file_index + 1
            ffile = get_next_file_and_close_current(file_list, file_index)
            # print file_list

    except NetlistParseError as npe:
        (msg,) = npe.args
        if len(msg):
            printing.print_general_error(msg)
        printing.print_parse_error(line_n, line)
        # if not read_netlist_from_stdin:
            # ffile.close()
        raise NetlistParseError(msg)

    # if not read_netlist_from_stdin:
        # ffile.close()

    models = parse_models(model_directives)

    # now we parse the subcircuits, we want a circuit.subckt object that holds the netlist code,
    # the nodes and the subckt name in a handy way.
    # We will create all the elements in the subckt every time it is
    # instantiated in the netlist file.
    subckts_dict = {}
    for subckt_temp in subckts_list_temp:
        subckt_obj = parse_sub_declaration(subckt_temp)
        if subckt_obj.name not in subckts_dict:
            subckts_dict.update({subckt_obj.name: subckt_obj})
        else:
            raise NetlistParseError("subckt " + \
                subckt_obj.name + " has been redefined")

    circ += main_netlist_parser(circ, netlist_lines, subckts_dict, models)
    circ.models = models
    return (circ, directives, postproc)


def main_netlist_parser(circ, netlist_lines, subckts_dict, models):
    elements = []
    parse_function = {
        'c': lambda line: parse_elem_capacitor(line, circ),
        'd': lambda line: parse_elem_diode(line, circ, models),
        'e': lambda line: parse_elem_vcvs(line, circ),
        'f': lambda line: parse_elem_cccs(line, circ),
        'g': lambda line: parse_elem_vccs(line, circ),
        'h': lambda line: parse_elem_ccvs(line, circ),
        'i': lambda line: parse_elem_isource(line, circ),
        'k': lambda line: parse_elem_inductor_coupling(line, circ, elements),
        'l': lambda line: parse_elem_inductor(line, circ),
        'm': lambda line: parse_elem_mos(line, circ, models),
        'r': lambda line: parse_elem_resistor(line, circ),
        's': lambda line: parse_elem_switch(line, circ, models),
        'v': lambda line: parse_elem_vsource(line, circ),
        'x': lambda line: parse_sub_instance(line, circ, subckts_dict, models),
        'y': lambda line: parse_elem_user_defined(line, circ)
    }
    try:
        for line, line_n in netlist_lines:
            # elements: detect the element type and call the
            # appropriate parsing function
            # we always use normal convention V opposite to I
            # n1 is +, n2 is -, current flows from + to -
            try:
                elements += parse_function[line[0]](line)
            except KeyError:
                raise NetlistParseError("Parser: do not know how to parse" +
                                        " '%s' elements." % line[0])
    #   Handle errors from individual parse functions
    except NetlistParseError as npe:
        (msg,) = npe.args
        if len(msg):
            printing.print_general_error(msg)
        printing.print_parse_error(line_n, line)
        raise NetlistParseError(msg)

    return elements


def get_next_file_and_close_current(file_list, file_index):
    if file_list[file_index - 1][2]:
        file_list[file_index - 1][0].close()
    if file_index == len(file_list):
        ffile = None
    else:
        ffile = open(file_list[file_index][1], "r")
        file_list[file_index][0] = ffile
    return ffile


def parse_models(models_lines):
    models = {}
    for line, line_n in models_lines:
        tokens = line.replace("(", "").replace(")", "").split()
        if len(tokens) < 3:
            raise NetlistParseError("parse_models(): syntax error in model" +
                                    " declaration on line " + str(line_n) +
                                    ".\n\t" + line)
        model_label = tokens[2]
        model_type = tokens[1]
        model_parameters = {}
        for index in range(3, len(tokens)):
            if tokens[index][0] == "*":
                break
            (label, value) = parse_param_value_from_string(tokens[index])
            model_parameters.update({label.upper(): value})
        if model_type == "ekv":
            model_iter = ekv.ekv_mos_model(**model_parameters)
            model_iter.name = model_label
        elif model_type == "mosq":
            model_iter = mosq.mosq_mos_model(**model_parameters)
            model_iter.name = model_label
        elif model_type == "diode" or model_type == 'd':
            model_parameters.update({'name': model_label})
            model_iter = diode.diode_model(**model_parameters)
        elif model_type == "sw":
            model_parameters.update({'name': model_label})
            model_iter = switch.vswitch_model(**model_parameters)
        # elif model_type == "csw":
        #   model_parameters.update({'name':model_label})
        #   model_iter = switch.iswitch_model(**model_parameters)
        else:
            raise NetlistParseError("parse_models(): Unknown model (" +
                                    model_type + ") on line " + str(line_n) +
                                    ".\n\t" + line,)
        models.update({model_label: model_iter})
    return models


def parse_elem_resistor(line, circ):
    """Parses a resistor from the line supplied, adds its nodes to the circuit
    instance circ and returns a list holding the resistor element.

    **Parameters:**

    line : string
        The netlist line.
    circ : circuit instance
        The circuit instance to which the resistor is to be connected.

    **Returns:**

    elements_list : list
        A list containing a :class:`ahkab.devices.Resistor` element.

    """
    line_elements = line.split()
    if len(line_elements) < 4 or (len(line_elements) > 4 and not line_elements[4][0] == "*"):
        raise NetlistParseError("parse_elem_resistor(): malformed line")

    ext_n1 = line_elements[1]
    ext_n2 = line_elements[2]
    n1 = circ.add_node(ext_n1)
    n2 = circ.add_node(ext_n2)

    value = convert_units(line_elements[3])

    if value == 0:
        raise NetlistParseError("parse_elem_resistor(): ZERO-valued resistors are not allowed.")

    elem = devices.Resistor(part_id=line_elements[0], n1=n1, n2=n2, value=value)

    return [elem]


def parse_elem_capacitor(line, circ):
    """Parses a capacitor from the line supplied, adds its nodes to the circuit
    instance circ and returns a list holding the capacitor element.

    **Parameters:**

    line : string
        The netlist line.
    circ : circuit instance
        The circuit to which the capacitor is to be connected.

    **Returns:**

    elements_list : list
        A list containing a :class:`ahkab.devices.Capacitor` element.
    """
    line_elements = line.split()
    if len(line_elements) < 4 or \
       (len(line_elements) > 5 and not line_elements[5][0] == "*" and
        not line_elements[4][0] == "*"):
        raise NetlistParseError("parse_elem_capacitor(): malformed line")

    ic = None
    if len(line_elements) == 5 and not line_elements[4][0] == '*':
        (label, value) = parse_param_value_from_string(line_elements[4])
        if label == "ic":
            ic = convert_units(value)
        else:
            raise NetlistParseError("parse_elem_capacitor(): unknown parameter " + label)

    ext_n1 = line_elements[1]
    ext_n2 = line_elements[2]
    n1 = circ.add_node(ext_n1)
    n2 = circ.add_node(ext_n2)

    elem = devices.Capacitor(part_id=line_elements[0], n1=n1, n2=n2,
                             value=convert_units(line_elements[3]), ic=ic)

    return [elem]


def parse_elem_inductor(line, circ):
    """Parses a inductor from the line supplied, adds its nodes to the circuit
    instance circ and returns a list holding the inductor element.

    **Parameters:**

    line : string
        The netlist line.
    circ : circuit instance
        The circuit to which the inductor is to be connected.

    **Returns:**

    elements_list : list
        A list containing a :class:`ahkab.devices.Inductor` element.
    """
    line_elements = line.split()
    if len(line_elements) < 4 or (len(line_elements) > 5 and not line_elements[6][0] == "*"):
        raise NetlistParseError("parse_elem_inductor(): malformed line")

    ic = None
    if len(line_elements) == 5 and not line_elements[4][0] == '*':
        (label, value) = parse_param_value_from_string(line_elements[4])
        if label == "ic":
            ic = convert_units(value)
        else:
            raise NetlistParseError("parse_elem_inductor(): unknown parameter " + label)

    ext_n1 = line_elements[1]
    ext_n2 = line_elements[2]
    n1 = circ.add_node(ext_n1)
    n2 = circ.add_node(ext_n2)

    elem = devices.Inductor(part_id=line_elements[0], n1=n1, n2=n2,
                            value=convert_units(line_elements[3]), ic=ic)

    return [elem]


def parse_elem_inductor_coupling(line, circ, elements=[]):
    """Parses a inductor coupling from the line supplied,
    returns a list holding the inductor coupling element.

    **Parameters:**

    line : string
        The netlist line.
    circ : circuit instance
        The circuit to which the inductor coupling is to be connected.

    **Returns:**

    elements_list : list
        A list containing a :class:`ahkab.devices.InductorCoupling` element.
    """
    line_elements = line.split()
    if len(line_elements) < 4 or (len(line_elements) > 4 and not line_elements[5][0] == "*"):
        raise NetlistParseError("parse_elem_inductor_coupling(): malformed line")

    part_id = line_elements[0]
    L1 = line_elements[1]
    L2 = line_elements[2]

    try:
        Kvalue = convert_units(line_elements[3])
    except ValueError:
        (label, value) = parse_param_value_from_string(line_elements[3])
        if not label == "k":
            raise NetlistParseError("parse_elem_inductor_coupling(): unknown parameter " + label)
        Kvalue = convert_units(value)

    L1elem, L2elem = None, None

    for e in elements:
        if isinstance(e, devices.Inductor) and L1 == e.part_id:
            L1elem = e
        elif isinstance(e, devices.Inductor) and L2 == e.part_id:
            L2elem = e

    if L1elem is None or L2elem is None:
        error_msg = "parse_elem_inductor_coupling(): One or more coupled" + \
                    " inductors for %s were not found: %s (found: %s), %s (found: %s)." % \
                    (part_id, L1, L1elem is not None, L2, L2elem is not None)
        raise NetlistParseError(error_msg)

    M = math.sqrt(L1elem.value * L2elem.value) * Kvalue

    elem = devices.InductorCoupling(part_id=part_id, L1=L1, L2=L2, K=Kvalue,
                                    M=M)
    L1elem.coupling_devices.append(elem)
    L2elem.coupling_devices.append(elem)

    return [elem]


def parse_elem_vsource(line, circ):
    """Parses a voltage source from the line supplied, adds its nodes to the
    circuit instance and returns a list holding the element.

    **Parameters:**

    line : string
        The netlist line.
    circ : circuit instance
        The circuit in which the voltage source is to be inserted.

    **Returns:**

    elements_list : list
        A list containing a :class:`ahkab.devices.VSource` element.
    """
    line_elements = line.split()
    if len(line_elements) < 3:
        raise NetlistParseError("parse_elem_vsource(): malformed line")

    dc_value = None
    vac = None
    function = None

    index = 3
    while True:  # for index in range(3, len(line_elements)):
        if index >= len(line_elements):
            break
        if line_elements[index][0] == '*':
            break

        label, value = parse_param_value_from_string(line_elements[index])

        if label == 'type':
            if value == 'vdc':
                param_number = 0
            elif value == 'vac':
                param_number = 0
            elif value == 'pulse':
                param_number = 7
            elif value == 'exp':
                param_number = 6
            elif value == 'sin':
                param_number = 5
            elif value == 'sffm':
                param_number = 5
            elif value == 'am':
                param_number = 5
            else:
                raise NetlistParseError("parse_elem_vsource(): unknown signal" +
                                        "type %s" % value)
            if param_number and function is None:
                function = parse_time_function(value,
                                               line_elements[index + 1:
                                                             index + param_number
                                                             + 1],
                                               "voltage")
                index = index + param_number
                # continue
            elif function is not None:
                raise NetlistParseError("parse_elem_vsource(): only a time function can be defined.")
        elif label == 'vdc':
            dc_value = convert_units(value)
        elif label == 'vac':
            vac = convert_units(value)
        else:
            raise NetlistParseError("parse_elem_vsource(): unknown type %s" %
                                    label)
        index = index + 1

    if dc_value == None and function == None:
        raise NetlistParseError("parse_elem_vsource(): neither vdc nor a time function are defined.")

    # usual
    ext_n1 = line_elements[1]
    ext_n2 = line_elements[2]
    n1 = circ.add_node(ext_n1)
    n2 = circ.add_node(ext_n2)

    elem = devices.VSource(part_id=line_elements[0], n1=n1, n2=n2,
                           dc_value=dc_value, ac_value=vac)

    if function is not None:
        elem.is_timedependent = True
        elem._time_function = function

    return [elem]


def parse_elem_isource(line, circ):
    """Parses a current source from the line supplied, adds its nodes to the
    circuit instance and returns a list holding the current source element.

    **Parameters:**

    line : string
        The netlist line.
    circ : circuit instance
        The circuit in which the current source is to be inserted.

    **Returns:**

    elements_list : list
        A list containing a :class:`ahkab.devices.ISource` element.
    """
    line_elements = line.split()
    if len(line_elements) < 3:
        raise NetlistParseError("parse_elem_isource(): malformed line")

    dc_value = None
    iac = None
    function = None

    index = 3
    while True:  # for index in range(3, len(line_elements)):
        if index == len(line_elements):
            break
        if line_elements[index][0] == '*':
            break

        (label, value) = parse_param_value_from_string(line_elements[index])

        if label == 'type':
            if value == 'idc':
                param_number = 0
            elif value == 'iac':
                param_number = 0
            elif value == 'pulse':
                param_number = 7
            elif value == 'exp':
                param_number = 6
            elif value == 'sin':
                param_number = 5
            elif value == 'sffm':
                param_number = 5
            elif value == 'am':
                param_number = 5
            else:
                raise NetlistParseError("parse_elem_isource(): unknown signal type.")
            if param_number and function is None:
                function = parse_time_function(value,
                                               line_elements[index + 1:
                                                             index + param_number + 1],
                                               "current")
                index = index + param_number
            elif function is not None:
                raise NetlistParseError("parse_elem_isource(): only a time function can be defined.")
        elif label == 'idc':
            dc_value = convert_units(value)
        elif label == 'iac':
            iac = convert_units(value)
        else:
            raise NetlistParseError("parse_elem_isource(): unknown type "+label)
        index = index + 1

    if dc_value == None and function == None:
        raise NetlistParseError("parse_elem_isource(): neither idc nor a time function are defined.")

    ext_n1 = line_elements[1]
    ext_n2 = line_elements[2]
    n1 = circ.add_node(ext_n1)
    n2 = circ.add_node(ext_n2)

    elem = devices.ISource(part_id=line_elements[0], n1=n1, n2=n2,
                           dc_value=dc_value, ac_value=iac)

    if function is not None:
        elem.is_timedependent = True
        elem._time_function = function

    return [elem]


def parse_elem_diode(line, circ, models=None):
    """Parses a diode from the line supplied, adds its nodes to the circuit
    instance and returns a list holding the diode element.

    Diode syntax:

    ::

        DX N+ N- <MODEL_LABEL> <AREA=xxx>

    **Parameters:**

    line : string
        The netlist line.
    circ : circuit instance
        The circuit in which the diode will be inserted.

    **Returns:**

    elements_list : list
        A list containing a :class:`ahkab.diode.Diode` element.
    """
    # sarebbe bello implementare anche: <IC=VD> <TEMP=T>
    Area = None
    T = None
    ic = None
    off = False

    line_elements = line.split()
    if len(line_elements) < 4:
        raise NetlistParseError("")

    model_label = line_elements[3]

    for index in range(4, len(line_elements)):
        if line_elements[index][0] == '*':
            break
        param, value = parse_param_value_from_string(line_elements[index])

        value = convert_units(value)
        if param == "area":
            Area = value
        elif param == "t":
            T = value
        elif param == "ic":
            ic = value
        elif param == "off":
            if not len(value):
                off = True
            else:
                off = convert_boolean(value)
        else:
            raise NetlistParseError("parse_elem_diode(): unknown parameter " + param)

    ext_n1 = line_elements[1]
    ext_n2 = line_elements[2]
    n1 = circ.add_node(ext_n1)
    n2 = circ.add_node(ext_n2)

    if model_label not in models:
        raise NetlistParseError("parse_elem_diode(): Unknown model id: " + model_label)
    elem = diode.diode(part_id=line_elements[0], n1=n1, n2=n2, model=models[
                       model_label], AREA=Area, ic=ic, off=off)
    return [elem]


def parse_elem_mos(line, circ, models):
    """Parses a MOS transistor from the line supplied, adds its nodes to the
    circuit instance and returns a list holding the element.

    MOS syntax:

    ::
        MX ND NG NS KP=xxx Vt=xxx W=xxx L=xxx type=n/p <LAMBDA=xxx>

    **Parameters:**

    line : string
        The netlist line.
    circ : circuit instance
        The circuit to which the element will be added.

    **Returns:**

    elements_list : list
        A list containing a MOS element.
    """
    #(self, nd, ng, ns, kp=1.0e-14, w, l, mos_type='n', lambd=0, type_of_elem="mosq")

    line_elements = line.split()
    if len(line_elements) < 6:
        raise NetlistParseError("parse_elem_mos(): required parameters are missing.")
        # print "MX ND NG NS model_id W=xxx L=xxx"

    model_label = line_elements[5]

    # kp = None
    w = None
    l = None
    # mos_type = None
    # vt = None
    m = 1
    n = 1
    # lambd = 0 # va is supposed infinite if not specified
    for index in range(6, len(line_elements)):
        if line_elements[index][0] == '*':
            break
        param, value = parse_param_value_from_string(line_elements[index])
        if param == "w":
            w = convert_units(value)
        elif param == "l":
            l = convert_units(value)
        elif param == "m":
            m = convert_units(value)
        elif param == "n":
           n = convert_units(value)
        else:
            raise NetlistParseError("parse_elem_mos(): unknown parameter " + param)

    if (w is None) or (l is None):
        raise NetlistParseError('parse_elem_mos(): required parameter ' +
                                'w'*(w is None) + ' and '*
                                (w is None and l is None) + 'l'*(l is None)+
                                'missing.')
        # print "MX ND NG NS W=xxx L=xxx <M=xxx> <N=xxx>"

    ext_nd = line_elements[1]
    ext_ng = line_elements[2]
    ext_ns = line_elements[3]
    ext_nb = line_elements[4]
    nd = circ.add_node(ext_nd)
    ng = circ.add_node(ext_ng)
    ns = circ.add_node(ext_ns)
    nb = circ.add_node(ext_nb)

    if model_label not in models:
        raise NetlistParseError("parse_elem_mos(): Unknown model ID: " + model_label)

    elem = None
    if isinstance(models[model_label], ekv.ekv_mos_model):
        elem = ekv.ekv_device(line_elements[0], nd, ng, ns, nb, w, l,
                              models[model_label], m, n)
    elif isinstance(models[model_label], mosq.mosq_mos_model):
        elem = mosq.mosq_device(line_elements[0], nd, ng, ns, nb, w, l,
                                models[model_label], m, n)
    else:
        raise NetlistParseError("parse_elem_mos(): Unknown MOS model type: " + model_label)

    return [elem]


def parse_elem_vcvs(line, circ):
    """Parses a voltage-controlled voltage source (VCVS) from the line
    supplied, adds its nodes to the circuit instance circ and returns a
    list holding the VCVS element.

    **Parameters:**

    line : string
        The netlist line.
    circ : circuit instance
        The circuit in which the VCVS is to be inserted.

    **Returns:**

    elements_list : list
        A list containing a :class:`ahkab.devices.EVSource` element.
    """
    line_elements = line.split()
    if len(line_elements) < 6 or (len(line_elements) > 6 and not line_elements[6][0] == "*"):
        raise NetlistParseError("")

    ext_n1 = line_elements[1]
    ext_n2 = line_elements[2]
    ext_sn1 = line_elements[3]
    ext_sn2 = line_elements[4]
    n1 = circ.add_node(ext_n1)
    n2 = circ.add_node(ext_n2)
    sn1 = circ.add_node(ext_sn1)
    sn2 = circ.add_node(ext_sn2)

    elem = devices.EVSource(part_id=line_elements[0], n1=n1, n2=n2, sn1=sn1,
                            sn2=sn2, value=convert_units(line_elements[5]))

    return [elem]


def parse_elem_ccvs(line, circ):
    """Parses a current-controlled voltage source (CCVS) from the line supplied,
    adds its nodes to the circuit instance and returns a list holding
    the CCVS element.

    CCVS syntax:

    ::

       HXXX N1 N2 VNAME VALUE


    **Parameters:**

    line : string
        The netlist line.
    circ : circuit instance
        The circuit in which the CCVS is to be inserted.

    **Returns:**

    elements_list : list
        A list containing a :class:`ahkab.devices.HVSource` element.
    """
    #  0    1  2  3     4
    line_elements = line.split()
    if len(line_elements) < 5 or (len(line_elements) > 5 and not
                                  line_elements[5][0] == "*"):
        raise NetlistParseError("")

    ext_n1 = line_elements[1]
    ext_n2 = line_elements[2]
    n1 = circ.add_node(ext_n1)
    n2 = circ.add_node(ext_n2)

    elem = devices.HVSource(part_id=line_elements[0], n1=n1, n2=n2,
                            source_id=line_elements[3],
                            value=convert_units(line_elements[4]))

    return [elem]


def parse_elem_vccs(line, circ):
    """Parses a voltage-controlled current source (VCCS) from the line
    supplied, adds its nodes to the circuit instance and returns a
    list holding the VCCS element.

    Syntax:

    ::

        GX N+ N- NC+ NC- VALUE

    **Parameters:**

    line : string
        The netlist line.
    circ : circuit instance
        The circuit in which the VCCS is to be inserted.

    **Returns:**

    elements_list : list
        A list containing a :class:`ahkab.devices.GISource` element.
    """
    line_elements = line.split()
    if len(line_elements) < 6 or (len(line_elements) > 6
       and not line_elements[6][0] == "*"):
        raise NetlistParseError("")

    ext_n1 = line_elements[1]
    ext_n2 = line_elements[2]
    ext_sn1 = line_elements[3]
    ext_sn2 = line_elements[4]
    n1 = circ.add_node(ext_n1)
    n2 = circ.add_node(ext_n2)
    sn1 = circ.add_node(ext_sn1)
    sn2 = circ.add_node(ext_sn2)

    elem = devices.GISource(part_id=line_elements[0], n1=n1, n2=n2, sn1=sn1,
                            sn2=sn2, value=convert_units(line_elements[5]))

    return [elem]


def parse_elem_cccs(line, circ):
    """Parses a current-controlled current source (CCCS) from the line
    supplied, adds its nodes to the circuit instance and returns a
    list holding the CCCS element.

    Syntax::

        FX N+ N- VNAME VALUE

    **Parameters:**

    line : string
        The netlist line.
    circ : circuit instance
        The circuit in which the CCCS is to be inserted.

    **Returns:**

    elements_list : list
        A list containing a :class:`ahkab.devices.FISource` element.
    """

    line_elements = line.split()
    if len(line_elements) < 5 or (len(line_elements) > 5
       and not line_elements[5][0] == "*"):
        raise NetlistParseError("")

    ext_n1 = line_elements[1]
    ext_n2 = line_elements[2]
    source_id = line_elements[3]
    n1 = circ.add_node(ext_n1)
    n2 = circ.add_node(ext_n2)

    elem = devices.FISource(part_id=line_elements[0], n1=n1, n2=n2,
                            source_id=source_id,
                            value=convert_units(line_elements[4]))

    return [elem]


def parse_elem_switch(line, circ, models=None):
    """Parses a switch device from the line supplied, adds its nodes to
    the circuit instance and returns a list holding the switch element.

    General syntax::

        SW1 n1 n2 ns1 ns2 model_label

    **Parameters:**

    line : string
        The netlist line.
    circ : circuit instance
        The circuit in which the switch is to be connected.
    models : dict, optional
        The currently defined models.

    **Returns:**

    elements_list : list
        A list containing a :class:`ahkab.switch.switch_device` element.
    """
    line_elements = line.split()
    if len(line_elements) < 6 or (len(line_elements) > 6 and not line_elements[6][0] == "*"):
        raise NetlistParseError("")

    ext_n1 = line_elements[1]
    ext_n2 = line_elements[2]
    ext_sn1 = line_elements[3]
    ext_sn2 = line_elements[4]
    n1 = circ.add_node(ext_n1)
    n2 = circ.add_node(ext_n2)
    sn1 = circ.add_node(ext_sn1)
    sn2 = circ.add_node(ext_sn2)

    model_label = line_elements[5]
    elem = None

    if model_label not in models:
        raise NetlistParseError("Unknown model id: " + model_label)
    if isinstance(models[model_label], switch.vswitch_model):
        elem = switch.switch_device(
            n1, n2, sn1, sn2, models[model_label], part_id=line_elements[0])
    else:
        raise NetlistParseError("Unknown MOS model type: " + model_label)

    return [elem]


def parse_elem_user_defined(line, circ):
    """Parses a user defined element.

    In order for this to work, you should write a module that supplies the
    elem class.

    Syntax:
    Y<X> <n1> <n2> module=<module_name> type=<type> [<param1>=<value1> ...]

    This method will attempt to load the module <module_name> and it will
    then look for a class named <type>.

    An object will be instatiated with the following arguments:
    n1, n2, param_dict, get_int_id_func, convert_units_func
    Where:
    n1: is the anode of the element
    n2: is the cathode
    param_dict: is a dictionary, its elements are {param1:value1, ...}
    get_int_id_func, convert_units_func are two function that may be used
    in the __init__ method, if needed.
    get_int_id_func: a function that gives back the internal name of a node
    convert_units_func: utility function to convert eg 1p -> 1e-12

    See ideal_oscillators.py for a reference implementation.
    **Parameters:**

    line : string
        The netlist line.
    circ : circuit instance.
        The circuit to which the element will be added.

    **Returns:**

    elements_list : list
        A list containing a :class:`ahkab.devices.HVSource` element.

    Parameters:
    line: the line
    circ: the circuit instance.

    Returns: [userdef_elem]
    """
    line_elements = line.split()

    if len(line_elements) < 4:
        raise NetlistParseError("")

    param_dict = {}
    for index in range(3, len(line_elements)):
        if line_elements[index][0] == '*':
            break

        param, value = parse_param_value_from_string(line_elements[index])

        if param not in param_dict:
            param_dict.update({param: value})
        else:
            raise NetlistParseError(param + " already defined.")

    if "module" in param_dict:
        module_name = param_dict.pop("module", None)
    else:
        raise NetlistParseError("module name is missing.")

    if module_name in circuit.user_defined_modules_dict:
        module = circuit.user_defined_modules_dict[module_name]
    else:
        try:
            fp, pathname, description = imp.find_module(module_name)
            module = imp.load_module(module_name, fp, pathname, description)
        except ImportError:
            raise NetlistParseError("module " + module_name + " not found.")
        circuit.user_defined_modules_dict.update({module_name: module})

    if "type" in param_dict:
        elem_type_name = param_dict.pop("type", None)
    else:
        raise NetlistParseError("type of element is missing.")

    try:
        elem_class = getattr(module, elem_type_name)
    except AttributeError:
        raise NetlistParseError("module doesn't have elem type: " + \
            elem_type_name)

    ext_n1 = line_elements[1]
    ext_n2 = line_elements[2]
    n1 = circ.add_node(ext_n1)
    n2 = circ.add_node(ext_n2)

    elem = elem_class(n1, n2, param_dict, circ.add_node,
                      convert_units, part_id=line_elements[0])

    selfcheck_result, error_msg = elem.check()
    if not selfcheck_result:
        raise NetlistParseError("module: " + module_name + " elem type: " + elem_type_name + " error: " +\
            error_msg)
    #   TODO fixme non so sicuro che sia una buona idea

    return [elem]


def parse_time_function(ftype, line_elements, stype):
    """Parses a time function of type ftype from the line_elements supplied.

    **Parameters:**

    ftype : str
        One among ``"pulse"``, ``"exp"``, ``"sin"``, ``"sffm"`` or ``"am"``.
    line_elements : list of strings
        The tokens describing the time function. The list mustn't hold the
        ``"type=<ftype>"`` element
    stype : str
        Set this to "current" for current sources, "voltage" for voltage sources

    See :class:`ahkab.time_functions.pulse`, :class:`ahkab.time_functions.sin`,
    :class:`ahkab.time_functions.exp`, :class:`ahkab.time_functions.sffm` and
    :class:`ahkab.time_functions.am` for more.

    **Returns:**

    time_function : object
        A time-function instance
    """
    if not ftype in time_fun_specs:
        raise NetlistParseError("Unknown time function: %s" % ftype)
    prot_params = list(copy.deepcopy(time_fun_specs[ftype]['tokens']))

    fun_params = {}
    for i in range(len(line_elements)):
        token = line_elements[i]
        if token[0] == "*":
            break
        if is_valid_value_param_string(token):
            (label, value) = token.split('=')
        else:
            label, value = None, token
        assigned = False
        for t in prot_params:
            if (label is None and t['pos'] == i) or label == t['label']:
                fun_params.update({t['dest']: convert(value, t['type'])})
                assigned = True
                break
        if assigned:
            prot_params.pop(prot_params.index(t))
            continue
        else:
            raise NetlistParseError("Unknown .%s parameter: pos %d (%s=)%s" % \
                                     (ftype.upper(), i, label, value))

    missing = []
    for t in prot_params:
        if t['needed']:
            missing.append(t['label'])
    if len(missing):
        raise NetlistParseError("%s: required parameters are missing: %s" % (ftype, " ".join(line_elements)))
    # load defaults for unsupplied parameters
    for t in prot_params:
        fun_params.update({t['dest']: t['default']})

    fun = time_functions[ftype](**fun_params)
    fun._type = "V" * \
        (stype.lower() == "voltage") + "I" * (stype.lower() == "current")
    return fun


def convert_units(string_value):
    """Converts a value conforming to SPICE's syntax to ``float``.

    Quote from the SPICE3 manual:

        A number field may be an integer field (eg 12, -44), a floating point
        field (3.14159), either an integer or a floating point number followed
        by an integer exponent (1e-14, 2.65e3), or either an integer or a
        floating point number followed by one of the following scale factors:

        T = 1e12, G = 1e9, Meg = 1e6, K = 1e3, mil = 25.4x1e-6, m = 1e-3, u =
        1e-6, n = 1e-9, p = 1e-12, f = 1e-15

    :raises ValueError: if the supplied string can't be interpreted according
    to the above.

    **Returns:**

    num : float
        A float representation of ``string_value``.
    """

    if type(string_value) is float:
        return string_value  # not actually a string!
    if not len(string_value):
        raise NetlistParseError("")

    index = 0
    string_value = string_value.strip().upper()
    while(True):
        if len(string_value) == index:
            break
        if not (string_value[index].isdigit() or string_value[index] == "." or
                string_value[index] == "+" or string_value[index] == "-" or
                string_value[index] == "E"):
            break
        index = index + 1
    if index == 0:
        # print string_value
        raise ValueError("Unable to parse value: %s" % string_value)
        # return 0
    numeric_value = float(string_value[:index])
    multiplier = string_value[index:]
    if len(multiplier) == 0:
        pass # return numeric_value
    elif multiplier == "T":
        numeric_value = numeric_value * 1e12
    elif multiplier == "G":
        numeric_value = numeric_value * 1e9
    elif multiplier == "K":
        numeric_value = numeric_value * 1e3
    elif multiplier == "M":
        numeric_value = numeric_value * 1e-3
    elif multiplier == "U":
        numeric_value = numeric_value * 1e-6
    elif multiplier == "N":
        numeric_value = numeric_value * 1e-9
    elif multiplier == "P":
        numeric_value = numeric_value * 1e-12
    elif multiplier == "F":
        numeric_value = numeric_value * 1e-15
    elif multiplier == "MEG":
        numeric_value = numeric_value * 1e6
    elif multiplier == "MIL":
        numeric_value = numeric_value * 25.4e-6
    else:
        raise ValueError("Unknown multiplier %s" % multiplier)
    return numeric_value


def parse_postproc(circ, postproc_direc):
    postproc_list = []
    for line, line_n in postproc_direc:
        if not line[0] == ".":
            continue

        try:
            line_elements = line.split()
            # plot
            if line_elements[0] == ".plot":
                plot_postproc = {}
                plot_postproc["type"] = "plot"
                plot_postproc["analysis"] = line_elements[1]
                if not (plot_postproc["analysis"] == "tran" or
                        plot_postproc["analysis"] == "pss" or
                        plot_postproc["analysis"] == "ac" or
                        plot_postproc["analysis"] == "dc"
                        ):
                    printing.print_general_error("Plotting is unsupported for" +
                                                 "analysis type " +
                                                 plot_postproc["analysis"])

                graph_labels = ""
                for glabel in line_elements[2:]:
                    graph_labels = graph_labels + " " + glabel

                l2l1 = plotting._split_netlist_label(graph_labels)

                if plot_postproc["analysis"] == "ac":
                    l2l1ac = []
                    for l2, l1 in l2l1:
                        if l1 is not None:
                            l1 = "|%s|" % (l1, )
                        else:
                            l1 = None
                        if l2 is not None:
                            l2 = "|%s|" % (l2, )
                        else:
                            l2 = None
                        l2l1ac.append((l2, l1))
                    l2l1 = l2l1ac
                plot_postproc["l2l1"] = l2l1
                postproc_list.append(plot_postproc)
            # fourier
            elif line_elements[0] == ".four":
                if type(convert_units(line_elements[1])) is not float:
                    raise NetlistParseError('postprocessing(): fourier' +
                                            ' fundamental \'%s\'?' %
                                            line_elements[1])
                fpv = {'type':'four',
                       'fund':convert_units(line_elements[1])}
                variables = []
                for le in line_elements[2:]:
                    if le[0] == '*':
                        break
                    variables += [plotting._split_netlist_label(le)[0]]
                fpv.update({'variables':tuple(variables)})
                postproc_list.append(fpv)
            elif line_elements[0] == '.fft':
                fpv = {'type': 'fft'}
                params = list(copy.deepcopy(fft_specs['fft']['tokens']))
                for i in range(len(line_elements[1:])):
                    token = line_elements[i + 1]
                    if token[0] == "*":
                        break
                    if is_valid_value_param_string(token):
                        label, value = token.split('=')
                    else:
                        label, value = None, token
                    assigned = False
                    for t in params:
                        if (label is None and t['pos'] == i) or label == t['label']:
                            fpv.update({t['dest']: convert(value, t['type'])})
                            assigned = True
                            break
                    if assigned:
                        params.pop(params.index(t))
                        continue
                    else:
                        raise NetlistParseError("Unknown .%s parameter: pos %d (%s=)%s" % \
                                                 (fpv[type].upper(), i, label, value))
                # is there anything required which is missing?
                missing = []
                for t in params:
                    if t['needed']:
                        missing.append(t['label'])
                if len(missing):
                    raise NetlistParseError("Required parameters are missing: %s" %
                                            (" ".join(line_elements)))
                # load defaults for unsupplied parameters
                for t in params:
                    fpv.update({t['dest']: t['default']})
                fpv.update({'label':
                            tuple(plotting._split_netlist_label(fpv['label'])[0])})
                postproc_list.append(fpv)
            else:
                raise NetlistParseError("Unknown postproc directive %s." %
                                        line_elements[0])
        except NetlistParseError as npe:
            (msg,) = npe.args
            if len(msg):
                printing.print_general_error(msg)
            printing.print_parse_error(line_n, line)
            raise NetlistParseError(msg)
    return postproc_list


def parse_ics(directives):
    ics = []
    for line, line_n in directives:
        if line[0] != '.':
            continue
        if line[:3] == '.ic':
            ics += [parse_ic_directive(line)]
    return ics


def parse_analysis(circ, directives):
    """Parses the analyses.

    **Parameters:**

    circ: circuit class instance
        The circuit description
    directives: list of tuples
        The list should be assembled as ``(line, line_number)``.

    Both of them are returned by ``parse_circuit()``

    **Returns:**

    a list of the analyses
    """
    an = []
    for line, line_n in directives:
        if line[0] != '.' or line[:3] == '.ic':
            continue
        line_elements = line.split()
        an += [parse_single_analysis(line)]
    return an


def parse_temp_directive(line):
    """Parses a TEMP directive:

    The syntax is::

        .TEMP <VALUE>

    """
    line_elements = line.split()
    for token in line_elements[1:]:
        if token[0] == "*":
            break
        value = convert_units(token)

    return {"type": "temp", "temp": value}


def parse_single_analysis(line):
    """Parses an analysis

    **Parameters:**

    line : str
        The netlist line from which an analysis statement is to be parsed.

    **Returns:**

    an : dict
        A dictionary with its parameters as keys.

    :raises NetlistParseError: if the analysis is not parsed correctly.
    """
    line_elements = line.split()
    an_type = line_elements[0].replace(".", "").lower()
    if not an_type in specs:
        raise NetlistParseError("Unknown directive: %s" % an_type)
    params = list(copy.deepcopy(specs[an_type]['tokens']))

    an = {'type': an_type}
    for i in range(len(line_elements[1:])):
        token = line_elements[i + 1]
        if token[0] == "*":
            break
        if is_valid_value_param_string(token):
            (label, value) = token.split('=')
        else:
            label, value = None, token
        assigned = False
        for t in params:
            if (label is None and t['pos'] == i) or label == t['label']:
                an.update({t['dest']: convert(value, t['type'])})
                assigned = True
                break
        if assigned:
            params.pop(params.index(t))
            continue
        else:
            raise NetlistParseError("Unknown .%s parameter: pos %d (%s=)%s" % \
                                     (an_type.upper(), i, label, value))

    missing = []
    for t in params:
        if t['needed']:
            missing.append(t['label'])
    if len(missing):
        raise NetlistParseError("Required parameters are missing: %s" %
                                (" ".join(line_elements)))
    # load defaults for unsupplied parameters
    for t in params:
        an.update({t['dest']: t['default']})

    # ad-hoc code for tran ... :(
    if an['type'] == 'tran':
        uic = int(an.pop('uic'))
        if uic == 0:
            an['x0'] = None
        elif uic == 1:
            an['x0'] = 'op'
        elif uic == 2:
            an['x0'] = 'op+ic'
        elif uic == 3:
            pass  # already set by ic_label
        else:
            raise NetlistParseError("Unknown UIC value: %d" % uic)
    # ... and pz :(
    if an['type'] == 'pz':
        an.update({'x0':'op'})

    return an


def is_valid_value_param_string(astr):
    """Has the string a form like ``<param_name>=<value>``?

    .. note::

        No spaces.

    **Returns:**

    ans : a boolean
        The answer to the above question.
    """
    work_astr = astr.strip()
    if work_astr.count("=") == 1:
        ret_value = True
    else:
        ret_value = False
    return ret_value


def convert(astr, rtype, raise_exception=False):
    """Convert a string to a different representation

    **Parameters:**

    astr : str
        The string to be converted.
    rtype : type
        One among ``float``, if a ``float`` sould be parsed from ``astr``,
        ``bool``, for parsing a boolean or ``str`` to get back a string (no
        parsing).
    raise_exception : boolean, optional
        Set this flag to ``True`` if you wish for this function to raise
        ``ValueError`` if parsing fails.

    **Returns:**

    ret : object
        The parsed data.
    """
    if rtype == float:
        try:
            ret = convert_units(astr)
        except ValueError as msg:
            if raise_exception:
                raise ValueError(msg)
            else:
                ret = astr
    elif rtype == str:
        ret = astr
    elif rtype == bool:
        ret = convert_boolean(astr)
    elif raise_exception:
        raise ValueError("Unknown type %s" % rtype)
    else:
        ret = astr
    return ret


def parse_param_value_from_string(astr, rtype=float, raise_exception=False):
    """Search the string for a ``<param>=<value>`` couple and returns a list.

    **Parameters:**

    astr : str
        The string to be converted.
    rtype : type
        One among ``float``, if a ``float`` sould be parsed from ``astr``,
        ``bool``, for parsing a boolean or ``str`` to get back a string (no
        parsing).
    raise_exception : boolean, optional
        Set this flag to ``True`` if you wish for this function to raise
        ``ValueError`` if parsing fails.

    **Returns:**

    ret : object
        The parsed data. If the conversion fails and ``raise_exception`` is not
        set, a ``string`` is returned.

    * If ``rtype`` is ``float`` (the type), its default value, the method will
      attempt converting ``astr`` to a float. If the conversion fails, a string
      is returned.
    * If set ``rtype`` to ``str`` (again, the type), a string will always be
      returned, as if the conversion failed.

    This prevents ``'0'`` (str) being detected as ``float`` and converted to 0.0,
    ending up being a new node instead of the reference.

    Notice that in ``<param>=<value>`` there is no space before or after the equal sign.

    **Returns:**

    alist : ``[param, value]``
        where ``param`` is a string and ``value`` is parsed as described.
    """
    if not is_valid_value_param_string(astr):
        return (astr, "")
    p, v = astr.strip().split("=")
    v = convert(v, rtype, raise_exception=False)
    return p, v


class NetlistParseError(Exception):
    """Netlist parsing exception."""
    pass


def convert_boolean(value):
    """Converts the following strings to a boolean:
    yes, 1, true to True
    no, false, 0 to False

    raises NetlistParserException

    Returns: boolean
    """
    if value == 'no' or value == 'false' or value == '0' or value == 0:
        return_value = False
    elif value == 'yes' or value == 'true' or value == '1' or value == 1:
        return_value = True
    else:
        raise NetlistParseError("invalid boolean: " + value)

    return return_value


def parse_ic_directive(line):
    """Parses an ic directive and assembles a dictionary accordingly.
    """
    line_elements = line.split()
    ic_dict = {}
    name = None
    for token in line_elements[1:]:
        if token[0] == "*":
            break

        (label, value) = parse_param_value_from_string(token)
        if label == "name" and name is None:
            name = value
            continue
        # the user should have specified either something like:
        # V(node)=10u
        # or something like:
        # I(Vtest)=100e-6
        ic_dict.update({label: convert_units(value)})
        # We may decide to check if the node exists and/or if the syntax
        # is correct and raise NetlistParseError if needed.

    if name is None:
        raise NetlistParseError("The 'name' parameter is missing")

    return {name: ic_dict}


def parse_sub_declaration(subckt_lines):
    """Returns a circuit.subckt instance that holds the subckt
    information, ready to be instantiated/called.
    """
    index = 0
    netlist_lines = []
    connected_nodes_list = []
    for line, line_n in subckt_lines:
        if index == 0:
            line_elements = line.split()
            if line_elements[0] != '.subckt':
                raise RuntimeError("BUG? parse_sub_declaration() \
                called on non-subckt text. (line" + str(line_n) + ")")
            name = line_elements[1]
            for node_name in line_elements[2:]:
                if node_name[0] == '0':
                    raise NetlistParseError("subckt " + name + \
                        " has a connection node named '0' (line" + str(
                            line_n) + ")")
                if node_name[0] == '*':
                    break
                else:
                    connected_nodes_list = connected_nodes_list + [node_name]
        else:
            netlist_lines = netlist_lines + [(line, "")]
        index = index + 1
    subck_inst = circuit.subckt(name, netlist_lines, connected_nodes_list)
    return subck_inst


def parse_sub_instance(line, circ, subckts_dict, models=None):
    """Parses a subckt call/instance.

    1. Gets name and nodes connections
    2. Looks in subckts_dict for a matching subckts_dict[name]
    3. Builds a circuit wrapper
    4. Calls main_netlist_parser() on the subcircuit code
       (with the wrapped circuit)

    Returns: a elements list

    """
    line_elements = line.split()
    if len(line_elements) < 2:
        raise NetlistParseError("")

    param_value_dict = {}
    name = None

    for index in range(1, len(line_elements)):
        if line_elements[index][0] == '*':
            break

        param, value = parse_param_value_from_string(line_elements[index],
                                                     rtype=str)
        param_value_dict.update({param:value})

    if "name" not in param_value_dict:
        raise NetlistParseError("missing 'name' in subckt call")
    if param_value_dict['name'] not in subckts_dict:
        raise NetlistParseError("subckt " + \
            param_value_dict['name'] + " is unknown")

    name = param_value_dict['name']
    subckt = subckts_dict[name]
    connection_nodes_dict = {}

    for param, value in param_value_dict.items():
        if param == 'name':
            continue
        if param in subckt.connected_nodes_list:
            connection_nodes_dict.update({param: value})
        else:
            raise NetlistParseError("unknown node " + param)

    # check all nodes are connected
    for node in subckt.connected_nodes_list:
        if node not in connection_nodes_dict:
            raise NetlistParseError("unconnected subckt node " + node)

    wrapped_circ = circuit._circuit_wrapper(circ, connection_nodes_dict,
                                            subckt.name, line_elements[0])

    elements_list = main_netlist_parser(wrapped_circ, subckt.code,
                                        subckts_dict, models)

    # Every subckt adds elements with the _same description_ (elem.part_id[1:])
    # We modify it so that each description is unique for every instance
    for element in elements_list:
        element.part_id = element.part_id[0] + \
                          "-" + wrapped_circ.prefix + element.part_id[1:]

    return elements_list


def parse_include_directive(line, netlist_wd):
    """.include <filename> [*comments]
    """
    line_elements = line.split()
    if not len(line_elements) > 1 or \
            (len(line_elements) > 2 and not line_elements[2][0] == '*'):
        raise NetlistParseError("")

    path = line_elements[1]
    if not os.path.isabs(path):
        # the user did not specify the full path.
        # the path is then assumed to be relative to the netlist location
        path = os.path.join(netlist_wd, path)
    if not utilities.check_file(path):
        raise RuntimeError("")

    fnew = open(path, "r")

    return [None, path, True]


def join_lines(fp, line):
    """Read the lines coming up in the file. Each line that starts with '+' is added to the
    previous line (line continuation rule). When a line not starting with '+' is found, the
    file is rolled back and the line is returned.
    """
    while True:
        last_pos = fp.tell()
        next = fp.readline()
        next = next.strip().lower()
        if not next:
            break
        elif next[0] == '+':
            line += ' ' + next[1:]
        else:
            fp.seek(last_pos)
            break
    return line



# fourier.py
# Module to perfor Fourier analysis of simulation data
# Copyright 2015 Giuseppe Venturini
#
# This file is part of the ahkab simulator.
#
# Ahkab is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# Ahkab is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License v2
# along with ahkab.  If not, see <http://www.gnu.org/licenses/>.

"""
This module offers the functions needed to perform a Fourier analysis of the
results of a simulation.

Module reference
''''''''''''''''

"""
from __future__ import (unicode_literals, absolute_import,
                        division, print_function)

import numpy as np
import numpy.fft as fft

from scipy.interpolate import InterpolatedUnivariateSpline
from scipy.signal import (bartlett, hann, hamming, blackman,
                          blackmanharris, gaussian, kaiser)

from . import options
from . import py3compat

# .FFT <output_var> <START = value> <STOP = value> <NP = value>
#+ <FORMAT = keyword> <WINDOW = keyword> <ALFA = value>
#+ <FREQ = value> <FMIN = value> <FMAX = value>
specs = {'fft':{'tokens':({
                          'label':None,
                          'pos':0,
                          'type':str,
                          'needed':True,
                          'dest':'label',
                          'default':None
                         },
                         {
                          'label':'start',
                          'pos':None,
                          'type':float,
                          'needed':False,
                          'dest':'stop',
                          'default':None
                         },
                         {
                          'label':'stop',
                          'pos':None,
                          'type':float,
                          'needed':False,
                          'dest':'start',
                          'default':None
                         },
                         {
                          'label':'from',
                          'pos':None,
                          'type':float,
                          'needed':False,
                          'dest':'from',
                          'default':None
                         },
                         {
                          'label':'to',
                          'pos':None,
                          'type':float,
                          'needed':False,
                          'dest':'to',
                          'default':None
                         },
                         {
                          'label':'np',
                          'pos':None,
                          'type':float,
                          'needed':False,
                          'dest':'np',
                          'default':0
                         },
                         {
                          'label':'format',
                          'pos':None,
                          'type':str,
                          'needed':False,
                          'dest':'uformat',
                          'default':'NORM'
                         },
                         {
                          'label':'window',
                          'pos':None,
                          'type':str,
                          'needed':False,
                          'dest':'window',
                          'default':options.RECT_WINDOW
                         },
                         {
                          'label':'alpha',
                          'pos':None,
                          'type':float,
                          'needed':False,
                          'dest':'alpha',
                          'default':3.
                         },
                         {
                          'label':'freq',
                          'pos':None,
                          'type':float,
                          'needed':False,
                          'dest':'freq',
                          'default':None
                         },
                         {
                          'label':'fmin',
                          'pos':None,
                          'type':float,
                          'needed':False,
                          'dest':'fmin',
                          'default':None
                         },
                         {
                          'label':'fmax',
                          'pos':None,
                          'type':float,
                          'needed':False,
                          'dest':'fmax',
                          'default':None
                         }
                        )
               }
           }

#.FOUR <freq> <x1> [<x2> [<x3 ...]]
def fourier(label, tran_results, fund):
    """Fourier analysis of the time evolution of a variable.

    In particular, the function uses the first 10 multiples of the fundamental
    frequency and a rectangular window.

    A variable amount of time data is used, resampled with a fixed
    time step. The length of the data is decided as follows:

    * The data should be taken from the end of the simulation, so that if there
      is any build-up or stabilization process, the Fourier analysis is not
      affected (or less affected) by it.
    * At least 1 period of the fundamental should be used.
    * Not more than 50% of the total simulation time should be used, if possible.
    * Respecting the above, as much data as possible should be used, as it leads
      to more accurate results.

    **Parameters:**

    label : str or tuple of str
        The identifier of a variable. Eg. ``'Vn1'`` or ``'I(VS)'``. If ``r`` is
        your ``tran_solution`` object, calling ``r.keys()`` will give you all
        the possible variable names for your result set.
        If a tuple of two identifiers is provided, the difference of the two,
        in the form ``label[0]-label[1]``, will be used.
    tran_results : tran_solution instance
        The TRAN results containing the time data for the ``'label'`` variable.
    fund : float
        The fundamental frequency, in Hertz.

    **Returns:**

    f : ndarray of floats
        The frequencies correspoding to the ``F`` array below.
    F : ndarray of complex data
        The result of the Fourier transform, including DC.
    THD : float
        The total harmonic distortion. This value, for a meaningful case, should
        be in the range (0, 1).
    """
    # we want a good number of periods. Not too big, to leave space for any
    # transient at the beginning, not too small, to have points in the FFT
    if type(label) in py3compat.string_types:
        data = tran_results[label]
    elif type(label) == tuple:
        if len(label) == 1 or (len(label) >= 2 and label[1] is None):
            data = tran_results[label[0]]
            label = label[0]
        else:
            data = tran_results[label[0]] - tran_results[label[1]]
            label = '%s - %s' % (label[0], label[1])
    ps = []
    for i in range(1, 10, 2):
        ps += [(tran_results.tstop - tran_results.tstart) // (i/fund)]
    ps = [i for i in ps if i > 0]
    nperiods = ps[int(np.ceil(len(ps)-1))]
    # setup start & stop accordingly
    start = tran_results.tstop - nperiods/fund
    stop = tran_results.tstop
    # 10 points, remember?
    sampling = 1./(2*10*fund)
    t = np.linspace(start, stop, (stop-start)//sampling, endpoint=False)
    idata = InterpolatedUnivariateSpline(tran_results.get_x(), data, k=2)
    f = fft.fftfreq(len(t), sampling)[::nperiods]
    F = fft.rfft(idata(t))[:-1:nperiods]
    f = f[:len(f)/2]
    THD = np.sqrt(sum(abs(F[2:])**2))/abs(F[1])
    return f, F, THD

# .FFT <output_var> <START = value> <STOP = value> <NP = value>
#+ <FORMAT = keyword> <WINDOW = keyword> <ALFA = value>
#+ <FREQ = value> <FMIN = value> <FMAX = value>
def spicefft(label, tran_results, freq=None, **args):
    """FFT analysis of the time evolution of a variable.

    This function is a much more flexible and complete version of the
    :func:`ahkab.fourier.fourier` function.

    The function uses a variable amount of time data, resampled with a fixed
    time step.
    The time interval is specified through the ``start`` and ``stop``
    parameters, if they are not set, all the available data is used.

    The function behaves differently whether the parameter ``freq`` is specified
    or not:

    * If the fundamental frequency ``freq`` (:math:`f` in the following) is
      specified, the function will perform an harmonic analysis, considering
      only the DC component and the harmonics of :math:`f` up to the 9th (ie
      :math:`f`, :math:`2f`, :math:`3f` :math:`\\dots` :math:`9f`).
    * If ``freq`` is left unspecified, a standard FFT analysis is performed,
      starting from :math:`f = 0`, to a frequency :math:`f_{max} =
      1/(2T_{TOT}n_p)`, where :math:`T_{TOT}` is the total length of the
      considered data in seconds and :math:`n_p` is the number of points in the
      FTT, set through the ``np`` parameter to this function.

    **Parameters:**

    label : str, or tuple of str
        The identifier of a variable. Eg. ``'Vn1'`` or ``'I(VS)'``. If ``r`` is
        your ``tran_solution`` object, calling ``r.keys()`` will give you all
        the possible variable names for your result set.
        If a tuple of two identifiers is provided, the difference of the two,
        in the form ``label[0]-label[1]``, will be used.
    tran_results : tran_solution instance
        The TRAN results containing the time data for the ``'label'`` variable.
    freq : float, optional
        The fundamental frequency, in Hertz. If it is specified, the output will
        be limited to the harmonics of this frequency. The THD evaluation will
        also be enabled.
    start : float, optional
        The first time instant to be considered for the transient analysis. If
        unspecified, it will be the beginning of the transient simulation.
    from : float, optional
        Alternative specification of the ``start`` parameter.
    stop : float, optional
        Last time instant to be considered for the FFT analysis. If unspecified,
        it will be the end time of the transient simulation.
    to : float, optional
        Alternative specification of the ``stop`` parameter.
    np : integer
        A power of two that specifies how many points should be used when
        computing the FFT. If it is set to a value that is not a power of 2, it
        will be rounded up to the nearest power of 2. It defaults to 1024. 
    window : str, optional
        The windowing type. The following values are available:

        * 'RECT' for a rectangular window, equivalent to no window at all.
        * 'BART', for a Bartlett window.
        * 'HANN', for a Hanning window.
        * 'HAMM' for a Hamming window.
        * 'BLACK' for a Blackman window.
        * 'HARRIS' for a Blackman-Harris window.
        * 'GAUSS' for a Gaussian window.
        * 'KAISER' for a Kaiser-Bessel window.

        The default is the rectangular window.
    alpha : float, optional
        The :math:`\\sigma` for a gaussian window or the :math:`beta` for a
        Kaiser window. Defaults to 3 and is ignored if a window different from
        Gaussian or Kaiser is selected.
    fmin : float, optional
        Suppress all data below this frequency, expressed in Hz. The suppressed
        data is neither returned nor used to compute the THD (if it is computed
        at all). The DC component is always preserved. Defaults to: return and
        use all data.
    fmax : float, optional
        The dual to ``fmin``, discard data above ``fmax`` and also do not use it
        if computing the THD. Defaults to infinity.

    **Returns:**

    f : ndarray of floats
        The frequencies, including the DC.
    F : ndarray of complex data
        The result of the Fourier transform, including DC.
    THD : float
        The total harmonic distortion, if ``freq`` was specified, ``None``
        otherwise.

    .. plot::

        import pylab
        import numpy as np
        import ahkab
        cir = ahkab.Circuit('Test FOUR and FFT')
        mys = ahkab.time_functions.sin(vo=0, va=1, freq=10e3)
        cir.add_resistor('R1', 'n1', cir.gnd, 1e3)
        cir.add_vsource('V1', 'n1', cir.gnd, 1, function=mys)
        tr = ahkab.new_tran(0, 1e-3, 1e-5, x0=None)
        r = ahkab.run(cir, tr)['tran']

        pylab.figure()
        pylab.title('Comparison of different windowing')
        for w in (ahkab.options.RECT_WINDOW, ahkab.options.BART_WINDOW,
                  ahkab.options.HANN_WINDOW, ahkab.options.HAMM_WINDOW,
                  ahkab.options.BLACK_WINDOW, ahkab.options.HARRIS_WINDOW,
                  ahkab.options.KAISER_WINDOW):
            fsp1, Fsp1, THDsp = ahkab.fourier.spicefft('vn1', r, window=w)
            pylab.plot(fsp1, 20*np.log10(abs(Fsp1)),
                       label=ahkab.options.WINDOWS_NAMES[w])
        pylab.xlabel('Frequency [Hz]'); pylab.ylabel('Magnitude [dB]')
        pylab.xlim((0, 100000))
        pylab.legend()

    """
    if type(label) in py3compat.string_types:
        data = tran_results[label]
    elif type(label) == tuple:
        if len(label) == 1 or (len(label) >= 2 and label[1] is None):
            data = tran_results[label[0]]
            label = label[0]
        else:
            data = tran_results[label[0]] - tran_results[label[1]]
            label = '%s - %s' % (label[0], label[1])
    if 'start' not in args and 'from' not in args:
        start = tran_results.tstart
    elif 'start' in args and args['start'] is not None:
        start = args['start']
    elif args['from'] is not None:
        start = args['from']
    else:
        start = tran_results.tstart

    if 'stop' not in args and 'to' not in args:
        stop = tran_results.tstop
    elif 'stop' in args and args['stop'] is not None:
        stop = args['stop']
    elif args['to'] is not None:
        stop = args['to']
    else:
        stop = tran_results.tstop

    if 'np' not in args:
        np2 = 1024
    else:
        # round to the nearest power of two
        np2 = 2 if args['np'] < 2 else 2**int(np.log2(args['np'] - 1) + 1)

    if 'window' in args:
        window_type = args['window'].upper()
        if window_type not in (options.RECT_WINDOW, options.BART_WINDOW,
                               options.HANN_WINDOW, options.HAMM_WINDOW,
                               options.BLACK_WINDOW, options.HARRIS_WINDOW,
                               options.GAUSS_WINDOW, options.KAISER_WINDOW):
            raise ValueError(('fft(): window may be %s, %s, %s, %s, %s, %s, %s'+
                              ' or %s, got %s') % (options.RECT_WINDOW,
                                                   options.BART_WINDOW,
                                                   options.HANN_WINDOW,
                                                   options.HAMM_WINDOW,
                                                   options.BLACK_WINDOW,
                                                   options.HARRIS_WINDOW,
                                                   options.GAUSS_WINDOW,
                                                   options.KAISER_WINDOW,
                                                   window_type))
    else:
        window_type = options.RECT_WINDOW
    alpha = args['alpha'] if 'alpha' in args else 3.
    fmin = args['fmin'] if 'fmin' in args else None
    fmax = args['fmax'] if 'fmax' in args else None

    if freq:
        nperiods = int((stop-start)*freq)
        # adjust start so we have an integer number
        # of periods
        start = stop - nperiods/freq
        sampling = 1./(2*freq*nperiods)
        t = np.linspace(start, stop, ((stop-start)//sampling + 1),
                        endpoint=False)
    else:
        sampling = (stop-start)/np2
        t = np.linspace(start, stop, np2, endpoint=False)
    idata = InterpolatedUnivariateSpline(tran_results.get_x(),
                                         data, k=2)
    window = {options.RECT_WINDOW: lambda x: 1.,
              options.BART_WINDOW: bartlett,
              options.HANN_WINDOW: hann,
              options.HAMM_WINDOW: hamming,
              options.BLACK_WINDOW: blackman,
              options.HARRIS_WINDOW: blackmanharris,
              options.GAUSS_WINDOW: lambda x: gaussian(x, std=alpha),
              options.KAISER_WINDOW: lambda x: kaiser(x, beta=alpha)}
    f = fft.fftfreq(len(t), sampling)
    f = f[:len(f)/2]
    F = fft.rfft(idata(t)*window[window_type](len(t)))[:-1]
    if freq:
        # downsample
        f = f[::nperiods]
        F = F[::nperiods]
    if fmin is not None and fmin > 0:
        if fmin > f.max():
            raise ValueError("fmin=%g Hz but max(f)=%g Hz" % (fmin, f.max()))
        # keep the DC
        F = np.concatenate((np.atleast_1d(F[0]), F[f >= fmin]))
        f = np.concatenate((np.atleast_1d(f[0]), f[f >= fmin]))
    if fmax is not None:
        if fmax < f[1:].min():
            raise ValueError("fmax=%g Hz but min(f)=%g Hz" %
                             (fmax, f[1:].min()))
        F = F[f <= fmax]
        f = f[f <= fmax]
    if freq:
        THD = np.sqrt(sum(abs(F[2:])**2))/abs(F[1])
    else:
        THD = None
    return f, F, THD




# -*- coding: iso-8859-1 -*-
# printing.py
# Printing module
# Copyright 2006 Giuseppe Venturini

# This file is part of the ahkab simulator.
#
# Ahkab is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# Ahkab is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License v2
# along with ahkab.  If not, see <http://www.gnu.org/licenses/>.

"""
This is the printing module of the simulator.
Using its functions, the output will be somewhat uniform.

The functions defined in this module can be divided in the following groups:

- :ref:`info-functions`: functions to print information, errors and warnings to
  the user during command-line execution.
- :ref:`netlist_syntax_printing`: functions to print conformingly to the netlist
  syntax, often to show information to the user for debugging purposes.
- :ref:`convenience_functions`: functions to abstract low level issues such as
  Unicode handling of text and number printing formats,
- :ref:`tabular_functions`: functions to format and display data into tables,
  which we provide straight from the ``tabulate`` module.
- :ref:`printing_analysis_results` in a consistent fashion.


.. _info-functions:

Informative functions
=====================

.. autosummary::
    print_general_error
    print_info_line
    print_parse_error
    print_warning

.. _netlist_syntax_printing:

Printing netlist lines
======================

.. autosummary::
    print_analysis

.. _printing_analysis_results:

Printing analysis results
=========================

.. autosummary::
    print_fourier
    print_spicefft
    print_symbolic_equations
    print_symbolic_results
    print_symbolic_transfer_functions

.. _convenience_functions:

Convenience functions
---------------------

.. autosummary::
    open_utf8
    printoptions

.. _tabular_functions:

Tabular formatting of data
--------------------------

.. autosummary::
    table

All functions in alphabetical order
===================================

"""

from __future__ import (unicode_literals, absolute_import,
                        division, print_function)

import contextlib
import sys
import os

import tabulate as _tabulate
import numpy as np

from . import options
from . import py3compat

if py3compat.PY2:
    import codecs
    if not py3compat.IPYTHON:
        UTF8Writer = codecs.getwriter('utf8')
        sys.stdout = UTF8Writer(sys.stdout)

def open_utf8(filename):
    """Get a file handle wrapped in a UTF-8 writer

    The file is opened in ``w`` mode.

    **Parameters:**

    filename : string
        The file name, just like you would pass to Python's built-in ``open()``
        method.

    **Returns:**

    fp : codecs.UTF8Writer object
        The wrapped file pointer.
    """
    if py3compat.PY2:
        fp = open(filename, 'w')
        UTF8Writer = codecs.getwriter('utf8')
        fp = UTF8Writer(fp)
    else:
        fp = open(filename, 'w', encoding='UTF-8')
    return fp

def print_analysis(an):
    """Prints an analysis to ``stdout`` in the netlist syntax

    **Parameters:**

    an : dict
        An analysis description in dictionary format.

    """
    if an["type"] == "op":
        print(".op", end="")
        for x in an:
            if x == 'type' or x == 'outfile' or x == 'verbose':
                continue
            print(" %s=%s" % (x, an[x]), end="")
        print("")
    elif an["type"] == "dc":
        print(".dc %(source)s start=%(start)g stop=%(stop)g step=%(step)g type=%(sweep_type)s" % an)
    elif an["type"] == "tran":
        sys.stdout.write(".tran tstep=" + str(an["tstep"]) + " tstop=" + str(
            an["tstop"]) + " tstart=" + str(an["tstart"]))
        if an["method"] is not None:
            print(" method=" + an["method"])
        else:
            print("")
    elif an["type"] == "shooting":
        sys.stdout.write(".shooting period=" + str(
            an["period"]) + " method=" + str(an["method"]))
        if an["points"] is not None:
            sys.stdout.write(" points=" + str(an["points"]))
        if an["step"] is not None:
            sys.stdout.write(" step=" + str(an["step"]))
        print(" autonomous=", an["autonomous"])


def print_general_error(description, print_to_stdout=False):
    """Prints an error message to ``stderr``

    **Parameters:**

    description : str
        The error description.

    print_to_stdout : bool, optional
        When set to ``True``, printing to ``stdout`` instead of ``stderr``.
        Defaults to ``False``.

    """
    the_error_message = "E: " + description
    if print_to_stdout:
        print(the_error_message)
    else:
        sys.stderr.write(the_error_message + "\n")
    return None


def print_warning(description, print_to_stdout=False):
    """Prints a warning message to ``stderr``

    **Parameters:**

    description : str
        The warning message.

    print_to_stdout : bool, optional
        When set to ``True``, printing to ``stdout`` instead of ``stderr``.
        Defaults to ``False``.

    """
    the_warning_message = "W: " + description
    if print_to_stdout:
        print(the_warning_message)
    else:
        sys.stderr.write(the_warning_message + "\n")
    return None


def print_info_line(msg_relevance_tuple, verbose, print_nl=True):
    """Conditionally print out a message

    **Parameters:**

    msg_relevance_tuple : sequence
        A tuple or list made of ``msg`` and ``importance``, where ``msg`` is a
        string, containing the information to be displayed to the user, and
        ``importance``, an integer, is its importance level. Zero corresponds to
        the highest possible importance level, which is always printed out by
        the simple algorithm discussed below.
    verbose : int
        The verbosity level of the program execution. Admissible levels are in
        the 0-6 range.
    print_nl : boolean, optional
        Whether a new line character should be appended or not to the string
        ``msg`` described above, if it's printed out. Defaults to ``True``.

    **Algorithm selecting when to print:**

    The message ``msg`` is printed out if the verbosity level is greater or
    equal than its importance.
    """
    msg, relevance = msg_relevance_tuple
    if verbose >= relevance:
        with printoptions(precision=options.print_precision,
                          suppress=options.print_suppress):
            if print_nl:
                print(msg)
            else:
                print(msg, end=' ')
                sys.stdout.flush()
    # else: suppressed.


def print_parse_error(nline, line, print_to_stdout=False):
    """Prints a parsing error to ``stderr``

    **Parameters:**

    nline : int
        The number of the line on which the error occurred.

    line : str
        The line of the file with the error.

    print_to_stdout : bool, optional
        When set to ``True``, printing to ``stdout`` instead of ``stderr``.
        Defaults to ``False``.

    """
    print_general_error(
        "Parse error on line " + str(nline) + ":", print_to_stdout)
    if print_to_stdout:
        print(line)
    else:
        sys.stderr.write(line + "\n")
    return None


def print_symbolic_results(x):
    """Print out symbolic results

    **Parameters:**

    x : dict
        A dictionary composed of elements like ``{v:expr}``,
        where ``v`` is a circuit variable and ``expr`` is the ``sympy``
        expression corresponding to it, as found by the solver.

    """
    keys = list(x.keys())
    keys.sort(lambda x, y: cmp(str(x), str(y)))
    for key in keys:
        print(str(key) + "\t = " + str(x[key]))


def print_symbolic_transfer_functions(x):
    """Print symbolic transfer functions

    **Parameters:**

    x : dict
        A dictionary of dictionaries. Each top level dictionary is a
        symbol : symbolic transfer function pair, eg. ``{vo/vin:<tf>}``.
        Each transfer function (``<tf>``) is itself a dictionary, having as keys
        the following strings: ``'gain'``, corresponding to the complete
        symbolic TF expression, ``'gain0'``, corresponding to the DC gain and
        ``'poles'`` and ``'zeros'``, corresponding to lists of symbolic
        expressions of the singularities.
    """
    keys = list(x.keys())
    keys.sort(lambda x, y: cmp(str(x), str(y)))
    for key in keys:
        print(str(key) + " = " + str(x[key]['gain']))
        print('\tDC: ' + str(x[key]['gain0']))
        for index in range(len(x[key]['poles'])):
            print('\tP' + str(index) + ":", str(x[key]['poles'][index]))
        for index in range(len(x[key]['zeros'])):
            print('\tZ' + str(index) + ":", str(x[key]['zeros'][index]))
    return None


def print_symbolic_equations(eq_list):
    """Print symbolic equations for visual inspection

    **Parameters:**

    eq_list : list
        The list of equations to be printed. This is what ``sympy`` will be
        asked to solve, typically.
    """
    print("+--")
    for eq in eq_list:
        print("| " + str(eq))
    print("+--")
    return

def print_fourier(label, f, F, THD, outfile='stdout'):
    """Print the results of a Fourier postprocess

    **Parameters:**

    label : str, or tuple of str
        The identifier of a variable. Eg. ``'Vn1'`` or ``'I(VS)'``.
        If a tuple of two identifiers is provided, it will be interpreted as the
        difference of the two, in the form ``label[0]-label[1]``.
    f : ndarray of floats
        The frequencies, including the DC.
    F : ndarray of complex data
        The result of the Fourier transform.
    THD : float, optional
        The total harmonic distortion, if ``freq`` was specified in the FFT
        analysis. Defaults to ``None``.
    outfile : str, optional
        The file name to print to. Defaults to ``'stdout'``, for the standard
        output.
     """
    if type(label) in py3compat.string_types:
        pass
    elif type(label) == tuple:
        if len(label) == 1 or (len(label) >= 2 and label[1] is None):
            label = label[0]
        else:
            label = '%s - %s' % (label[0], label[1])
    if outfile == 'stdout':
        fp = sys.stdout
    else:
        fp = open(outfile, 'w')
    fp.write('Fourier components of transient response %s\n' % label)
    DC = np.real_if_close(F[0])
    fp.write('DC component =  %g\n' % DC)
    headers = ('#', 'Frequency [Hz]', 'Magnitude', 'Normalized', 'Phase [deg]',
               'Normalized [deg]')
    ffttable = []
    i = np.argmax(np.abs(F))
    ref_f = f[i]
    ref_amplitude = np.abs(F)[i]
    ref_angle = np.angle(F, deg=True)[i]
    fp.write('Normalization to f = %g Hz, |A| = %g, phase(A) = %g deg\n' %
             (ref_f, ref_amplitude, ref_angle))
    for i in range(1, len(F)):
        ffttable += [[i, f[i], np.abs(F[i]), abs(F[i])/ref_amplitude,
                      np.angle(F[i], deg=True),
                      np.angle(F[i], deg=True) - ref_angle]]
    fp.write(table(ffttable, headers=headers))
    fp.write('\nTotal harmonic distortion: %g %%\n' % (100*THD))
    if outfile != 'stdout':
        fp.close()

def print_spicefft(label, f, F, THD=None, uformat='NORM', window=None,
                   outfile='stdout'):
    """Print the results of an FFT postprocess

    **Parameters:**

    label : str, or tuple of string
        The identifier of a variable. Eg. ``'Vn1'`` or ``'I(VS)'``.
        If a tuple of two identifiers is provided, it will be interpreted as the
        difference of the two, in the form ``label[0]-label[1]``.
    f : ndarray of floats
        The frequencies, including the DC.
    F : ndarray of complex data
        The result of the Fourier transform.
    THD : float, optional
        The total harmonic distortion, if ``freq`` was specified in the FFT
        analysis. Defaults to ``None``.
    uformat : str, optional
        The parameter format selects whether normalized or unnormalized
        magnitudes are printed. It is to be set to 'NORM' (default value) for
        normalized magnitude, to 'UNORM' for unnormalized.
    window : str, optional
        The window employed in the FFT analisys. Defaults to rectangular.
    outfile : str
        The file name to print to. Defaults to ``'stdout'``, for the standard
        output.
    """
    if type(label) in py3compat.string_types:
        pass
    elif type(label) == tuple:
        if len(label) == 1 or (len(label) >= 2 and label[1] is None):
            label = label[0]
        else:
            label = '%s - %s' % (label[0], label[1])
    if uformat.upper() not in ('NORM', 'UNORM'):
        raise ValueError('fft(): format may be "NORM" or "UNORM", got %s' %
                         uformat)
    angle = np.angle(F, deg=True)
    F = 20*np.log10(abs(F))  # dB
    if uformat.upper() == 'NORM':
        i = np.argmax(F)
        fnorm = f[i]
        Fnorm = F[i]
        F -= F.max()
        unit = 'dBc'
    else:
        unit = 'dB'
    if window is not None:
        window = window.upper()
        if window not in (options.RECT_WINDOW, options.BART_WINDOW,
                               options.HANN_WINDOW, options.HAMM_WINDOW,
                               options.BLACK_WINDOW, options.HARRIS_WINDOW,
                               options.GAUSS_WINDOW, options.KAISER_WINDOW):
            raise ValueError(('fft(): window may be %s, %s, %s, %s, %s, %s, %s'+
                              ' or %s, got %s') % (options.RECT_WINDOW,
                                                   options.BART_WINDOW,
                                                   options.HANN_WINDOW,
                                                   options.HAMM_WINDOW,
                                                   options.BLACK_WINDOW,
                                                   options.HARRIS_WINDOW,
                                                   options.GAUSS_WINDOW,
                                                   options.KAISER_WINDOW,
                                                   window))
    else:
        window = options.RECT_WINDOW
    # set up the output file
    if outfile == 'stdout':
        fp = sys.stdout
    else:
        fp = open(outfile, 'w')
    fp.write('FFT components of transient response %s\n' % label)
    fp.write('Window: %s\n' % options.WINDOWS_NAMES[window.upper()])
    fp.write('Start Frequency: %g Hz\n' % f[1])
    fp.write('Stop Frequency: %g Hz\n' % f[-1])
    if uformat.upper() == 'NORM':
        fp.write(('Normalization to carrier at %g Hz: '% fnorm) +
                 ('magnitude %g dB\n' % Fnorm))
    fp.write(('DC component: mag = %g ' + unit + '\n') % F[0])
    headers = ('#', 'Frequency [Hz]', 'Magnitude ['+unit+']', 'Phase [deg]')
    ffttable = []
    for i in range(1, len(F)):
        ffttable += [[i, f[i], F[i], angle[i]]]
    fp.write(table(ffttable, headers=headers))
    fp.write('\n')
    if THD is not None:
        fp.write('Total harmonic distortion: %g %%\n' % (100*THD))
    if outfile != 'stdout':
        fp.close()

def print_result_check(badvars, verbose=2):
    """Prints out the results of an OP check

    It assumes one set of results is calculated with :math:`G_{min}`, the other
    without.

    **Parameters:**

    badvars : list
        The list returned by :func:`results.op_solution.gmin_check`.
    verbose : int, optional
        The verbosity level, from 0 (silent) to 6.

    """
    if len(badvars):
        print("Warning: solution is heavvily dependent on gmin.")
        print("Affected variables:")
        for bv in badvars:
            print(bv)
    else:
        if verbose:
            print("Difference check within margins.")
            print("(Voltage: er=" + str(options.ver) + ", ea=" + str(options.vea) + \
                ", Current: er=" + \
                str(options.ier) + ", ea=" + str(options.iea) + ")")
    return None


def table(data, *args, **argsd):
    """Format a fixed width table for pretty printing

    No data processing is done here, instead we call `tabulate
    <https://pypi.python.org/pypi/tabulate/>`_'s ``tabulate.tabulate()``,
    passing all arguments unmodified.

    **Parameters:**

    data : list-of-lists or a dictionary of iterables or a 2D NumPy array (or more).
        The tabular data.

    The remaining arguments, not documented here, are:

    headers : sequence, optional
        An explicit list of column headers.
    tablefmt : str, optional
        Table formatting specification.
    floatfmt : str, optional
        Floats formatting specification.
    numalign : str, optional
        Alignment flag for numbers.
    stralign : str, optional
        Alignment specification for strings, eg. "right".
    missingval : str, optional
        Element for the missing values.

    """
    return _tabulate.tabulate(data, *args, **argsd)

@contextlib.contextmanager
def printoptions(*args, **kwargs):
    """A context manager for ``numpy.set_printoptions``"""
    original = np.get_printoptions()
    np.set_printoptions(*args, **kwargs)
    yield
    np.set_printoptions(**original)

locale = os.getenv('LANG')
if not locale:
    print_warning('Locale appears not set! please export LANG="en_US.UTF-8" or'
                  ' equivalent, ')
    print_warning('or ahkab\'s unicode support is broken.')

